import bpy
from ase.build import molecule, fcc111
from batoms.render import Render
from batoms import Batoms

try:
    from _common_helpers import has_display, set_cycles_res

    use_cycles = not has_display()
except ImportError:
    use_cycles = False

extras = dict(engine="cycles") if use_cycles else {}


def test_render():
    bpy.ops.batoms.delete()
    render = Render("test")
    assert isinstance(render, Render)
    render.studiolight = "paint.sl"
    render.viewport = [1, 0, 0]
    render.resolution = [200, 200]
    # from collection
    render = Render("test")
    assert isinstance(render, Render)
    assert render.studiolight == "paint.sl"
    assert render.resolution == [200, 200]


def test_render_gpu(h2o):
    h2o.render.init()
    h2o.render.gpu = True


def test_render_init(h2o):
    h2o.render.init()
    h2o.render.viewport = [1, 0, 0]
    h2o.render.resolution = [200, 200]
    if use_cycles:
        set_cycles_res(h2o)
    h2o.get_image(**extras)


def test_render_setter(h2o):
    h2o.render.init()
    h2o.render.viewport = [1, 0, 0]
    nh3 = molecule("NH3")
    nh3 = Batoms("nh3", from_ase=nh3)
    nh3.translate([3, 3, 0])
    nh3.render = Render("nh3")
    nh3.render.viewport = [0, 1, 0]
    nh3.render.lights.add("1", type="SUN", direction=[1, 0, 0])
    nh3.render.resolution = [200, 200]
    if use_cycles:
        set_cycles_res(nh3)
    nh3.get_image(output="render-setter.png", **extras)


def test_render_au111():
    bpy.ops.batoms.delete()
    atoms = fcc111("Au", size=(4, 4, 4), vacuum=0)
    au111 = Batoms(label="au111", from_ase=atoms)
    au111.cell[2, 2] += 5
    au111.render.resolution = [200, 200]
    if use_cycles:
        set_cycles_res(au111)
    au111.get_image(**extras)
    au111.render.engine = "eevee"
    au111.get_image([1, 1, 1], output="au111-eevee.png", **extras)
    au111.render.engine = "cycles"
    au111.get_image([1, 1, 1], output="au111-cycles.png", **extras)



"""
update batoms
"""

import bpy
import subprocess
import os
import logging

logger = logging.getLogger(__name__)

account_name = "beautiful-atoms"
repo_name = "beautiful-atoms"
DEFAULT_PLUGIN_NAME = "batoms"

repo_git = f"https://github.com/{account_name}/{repo_name}.git"  # noqa E231


def subprocess_run(cmds):
    try:
        p = subprocess.Popen(cmds, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        stdout, stderr = p.communicate()
        if p.returncode == 0:
            return True
        else:
            logger.critical(stdout, stderr)
    except (OSError, Exception) as exception:
        logger.critical(exception)
        return False
    return False


def has_git():
    return subprocess_run(["git", "--version"])


def gitclone(workdir=".", version="main", url=repo_git):
    """Make a git clone to the directory
    version can be a branch name or tag name
    """
    import shutil
    import pathlib

    batoms_dir = os.path.dirname(pathlib.Path(__file__).parent.resolve())
    addon_dir = os.path.dirname(batoms_dir)
    clone_into = os.path.join(addon_dir, repo_name)
    if os.path.exists(clone_into) and os.path.isdir(clone_into):
        shutil.rmtree(clone_into)
    commands = [
        "git",
        "clone",
        "--depth",
        "1",
        "-b",
        f"{version}",
        f"{url}",
        clone_into,
    ]
    flag = subprocess_run(commands)
    flag = True
    if flag:
        logger.info(f"Cloned repo into directory {clone_into}")
    else:
        logger.warning("Faild to clone")
    #
    src = os.path.join(clone_into, "batoms")
    if os.path.exists(batoms_dir) and os.path.isdir(batoms_dir):
        logger.info("Remove old Batoms folder {}".format(batoms_dir))
        shutil.rmtree(batoms_dir)
    logger.info("Move {} to {}".format(src, batoms_dir))
    shutil.move(src, batoms_dir)


class BatomsUpdateButton(bpy.types.Operator):
    """Update Batoms"""

    bl_idname = "batoms.update"
    bl_label = "Update Batoms"
    bl_description = "Update to the latest development version."

    def execute(self, context):
        if not has_git():
            self.report({"ERROR"}, "Please install Git first.")
            logger.warning("Please install Git first.")
            return {"CANCELLED"}
        gitclone()
        self.report({"INFO"}, "Update to the latest version successfully!")
        return {"FINISHED"}



import bpy
from bpy.props import (
    StringProperty,
    BoolProperty,
    IntProperty,
    FloatProperty,
    FloatVectorProperty,
    CollectionProperty,
)


class LightSetting(bpy.types.PropertyGroup):
    """ """

    flag: BoolProperty(name="flag", default=False)
    label: StringProperty(name="label", default="X")
    name: StringProperty(name="name", default="X")
    type: StringProperty(name="type", default="SUN")
    strength: FloatProperty(name="strength", soft_min=0.01, soft_max=1, default=0.01)
    lock_to_camera: BoolProperty(name="lock_to_camera", default=False)
    direction: FloatVectorProperty(name="direction", default=[0, 0, 1], size=3)
    look_at: FloatVectorProperty(name="look_at", default=[0, 0, 0], size=3)


class CameraSetting(bpy.types.PropertyGroup):
    """ """

    flag: BoolProperty(name="flag", default=False)
    label: StringProperty(name="label", default="X")
    name: StringProperty(name="name", default="X")
    type: StringProperty(name="type", default="ORTHO")
    direction: FloatVectorProperty(name="direction", default=[0, 0, 1], size=3)
    look_at: FloatVectorProperty(name="look_at", default=[0, 0, 0], size=3)


class Render(bpy.types.PropertyGroup):
    """This module defines the Render properties to extend
    Blender’s internal data.

    """

    flag: BoolProperty(name="flag", default=False)
    label: StringProperty(name="label", default="X")
    engine: StringProperty(name="engine", default="BLENDER_EEVEE")
    compute_device_type: StringProperty(name="compute_device_type", default="CUDA")
    animation: BoolProperty(name="animation", default=False)
    run_render: BoolProperty(name="run_render", default=True)
    gpu: BoolProperty(name="gpu", default=False)
    viewport: FloatVectorProperty(name="viewport", default=[0, 0, 1], size=3)
    center: FloatVectorProperty(name="center", default=[0, 0, 1], size=3)
    distance: FloatProperty(
        name="distance", description="Distance from camera", default=-1
    )
    padding: FloatVectorProperty(name="padding", default=[1, 1, 1, 1], size=4)

    light_ui_list_index: IntProperty(name="light_ui_list_index", default=0)
    camera_ui_list_index: IntProperty(name="camera_ui_list_index", default=0)
    # colleciton properties
    light_settings: CollectionProperty(name="lightsetting", type=LightSetting)
    camera_settings: CollectionProperty(name="camerasetting", type=CameraSetting)



import bpy
import pytest
from time import time
from batoms.pdbparser import read_pdb
from batoms import Batoms

try:
    from _common_helpers import has_display, set_cycles_res

    use_cycles = not has_display()
except ImportError:
    use_cycles = False

extras = dict(engine="cycles") if use_cycles else {}


def test_SAS(h2o):
    """ """
    h2o.molecular_surface.draw()
    assert len(h2o.molecular_surface.settings) == 1
    h2o.molecular_surface.settings["2"] = {"type": "SAS"}
    assert len(h2o.molecular_surface.settings) == 2
    # area = h2o.molecular_surface.get_psasa()


def test_SAS_location(h2o):
    """ """
    h2o.translate([10, 0, 0])
    h2o.molecular_surface.settings["1"] = {"type": "SAS"}
    h2o.molecular_surface.draw()
    # should be local positions
    assert bpy.data.objects["h2o_1_sas"].data.vertices[0].co[0] < 0
    # area = h2o.molecular_surface.get_psasa()


def test_EPM(c2h6so):
    """Electrostatic Potential Maps"""
    pytest.importorskip("openbabel")

    c2h6so.molecular_surface.draw()
    assert (
        "Attribute"
        not in bpy.data.objects["c2h6so_1_sas"].data.materials[0].node_tree.nodes
    )
    # color by potential
    c2h6so.molecular_surface.settings["1"].color_by = "Electrostatic_Potential"
    c2h6so.molecular_surface.draw()
    assert (
        "Attribute"
        in bpy.data.objects["c2h6so_1_sas"].data.materials[0].node_tree.nodes
    )
    if use_cycles:
        set_cycles_res(c2h6so)
    else:
        c2h6so.render.resolution = [200, 200]
    c2h6so.get_image([1, 0, 0], output="c2h6so-EMP.png", **extras)


def test_colored_by_volumetric_datas():
    """Electrostatic Potential Maps"""
    from batoms import Batoms
    from ase.io.cube import read_cube_data

    hartree, atoms = read_cube_data("../tests/datas/h2o-hartree.cube")
    bpy.ops.batoms.delete()
    h2o = Batoms("h2o", from_ase=atoms)
    h2o.volumetric_data["hartree"] = -hartree
    h2o.molecular_surface.draw()
    assert (
        "Attribute"
        not in bpy.data.objects["h2o_1_sas"].data.materials[0].node_tree.nodes
    )
    # color by potential
    h2o.molecular_surface.settings["1"].color_by = "hartree"
    h2o.molecular_surface.draw()
    assert (
        "Attribute" in bpy.data.objects["h2o_1_sas"].data.materials[0].node_tree.nodes
    )
    if use_cycles:
        set_cycles_res(h2o)
    else:
        h2o.render.resolution = [200, 200]
    h2o.get_image(
        [1, 0, 0], output="molecular_surface_colored_by_volumetric_datas.png", **extras
    )
    # teardown for next test
    bpy.ops.batoms.delete()


def test_SES(h2o):
    """ """
    h2o.molecular_surface.settings["1"].type = "SES"
    h2o.molecular_surface.draw()
    # area = h2o.molecular_surface.get_sasa(partial=True)[0]


def test_molecular_surface_ops(c2h6so):
    c2h6so.selects.add("1", [2, 4, 6, 7])
    c2h6so.selects.add("2", [3, 5, 8, 9])
    bpy.context.view_layer.objects.active = c2h6so.obj
    bpy.ops.surface.molecular_surface_add(name="2")
    assert len(c2h6so.molecular_surface.settings) == 2
    bpy.ops.surface.molecular_surface_remove(name="1")
    assert len(c2h6so.molecular_surface.settings) == 1
    print(c2h6so.molecular_surface.settings)
    bpy.ops.surface.molecular_surface_add(name="1")
    assert len(c2h6so.molecular_surface.settings) == 2
    c2h6so.molecular_surface.settings["1"].select = "1"
    c2h6so.molecular_surface.settings["2"].select = "2"
    c2h6so.molecular_surface.settings["2"].type = "SES"
    bpy.ops.surface.molecular_surface_draw()


def test_molecule_surface_uilist(ch4):
    """molecule_surface_uilist panel"""
    bpy.context.view_layer.objects.active = ch4.obj
    assert ch4.coll.Bmolecularsurface.ui_list_index == 0
    bpy.ops.surface.molecular_surface_add(name="2")
    assert ch4.coll.Bmolecularsurface.ui_list_index == 1


def test_SAS_protein():
    """ """
    bpy.ops.batoms.delete()
    prot = read_pdb("../tests/datas/1ema.pdb")  # 1tim
    prot = Batoms("1ema", from_ase=prot)
    prot.molecular_surface.settings["1"] = {"resolution": 0.4}
    prot.molecular_surface.draw()
    area = prot.molecular_surface.get_sasa("1")
    # area = prot.molecular_surface.get_psasa()
    assert abs(area[0] - 14461) < 10000
    prot.selects.add("A", "chain A")
    prot.molecular_surface.settings.add(
        "2", {"select": "A", "color": [0.8, 0.1, 0.1, 1.0]}
    )
    prot.molecular_surface.draw()


def test_SES_protein():
    """
    compare with jmol, QuickSES, EDTSurf
    isosurface resolution 3 molecular 1.4
    isosurface area set 0
            area                    volume
    """
    bpy.ops.batoms.delete()
    prot = read_pdb("../tests/datas/1ema.pdb")
    prot = Batoms("prot", from_ase=prot)
    tstart = time()
    prot.molecular_surface.settings["1"].type = "SES"
    prot.molecular_surface.draw()
    t = time() - tstart
    assert t < 10
    area = prot.molecular_surface.get_sesa("1")[0]
    assert abs(area - 8011) < 1000



import bpy
import bmesh
from bpy.types import Operator
from bpy.props import (
    BoolProperty,
    EnumProperty,
    IntProperty,
    StringProperty,
)
from batoms import Batoms
from batoms.ops.base import OperatorBatoms


class BondPairAdd(OperatorBatoms):
    bl_idname = "bond.bond_pair_add"
    bl_label = "Add Bond Pair"
    bl_description = "Add Bond Pair to a Batoms"

    species1: StringProperty(
        name="species", default="", description="Replaced by this species"
    )

    species2: StringProperty(
        name="species", default="", description="Replaced by this species"
    )

    def execute(self, context):
        if self.species1 == "" or self.species2 == "":
            return {"FINISHED"}
        obj = context.object
        batoms = Batoms(label=context.object.batoms.label)
        pair = (self.species1, self.species2)
        batoms.bond.settings.add(pair)
        context.view_layer.objects.active = obj
        self.report(
            {"INFO"}, "Add bond pair {} {}".format(self.species1, self.species2)
        )
        return {"FINISHED"}


class BondPairRemove(OperatorBatoms):
    bl_idname = "bond.bond_pair_remove"
    bl_label = "Remove Bond Pair"
    bl_description = "Remove Bond Pair to a Batoms"

    name: StringProperty(
        name="name", default="C-H", description="Name of bond pair to be removed"
    )

    all: BoolProperty(name="all", default=False, description="Remove all bond pairs")

    def execute(self, context):
        obj = context.object
        batoms = Batoms(label=obj.batoms.label)
        batoms.bond.settings.remove((self.name))
        context.view_layer.objects.active = obj
        return {"FINISHED"}


class BondDraw(OperatorBatoms):
    bl_idname = "bond.draw"
    bl_label = "Draw Bond"
    bl_description = "Draw Bond"

    model_style: EnumProperty(
        name="model_style",
        description="Structural models",
        items=(
            ("0", "Space-filling", "Use ball"),
            ("1", "Ball-and-stick", "Use ball and stick"),
            ("2", "Polyhedral", "Use polyhedral"),
            ("3", "Wireframe", "Use wireframe"),
        ),
        default="1",
    )

    def execute(self, context):
        obj = context.object
        batoms = Batoms(label=obj.batoms.label)
        batoms.draw(self.model_style)
        context.view_layer.objects.active = batoms.obj
        return {"FINISHED"}


class BondModify(Operator):
    bl_idname = "bond.bond_modify"
    bl_label = "Modify bond"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Modify bond"

    key: StringProperty(
        name="key", default="style", description="Replaced by this species"
    )

    style: StringProperty(name="style", default="0", description="style")

    show: BoolProperty(
        name="show",
        default=False,
    )
    order: IntProperty(
        name="Bond order",
        default=1,
    )

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj:
            return obj.batoms.type == "BOND" and obj.mode == "EDIT"
        else:
            return False

    def execute(self, context):
        obj = context.object
        data = obj.data
        bm = bmesh.from_edit_mesh(data)
        v = [s.index for s in bm.select_history if isinstance(s, bmesh.types.BMVert)]
        batoms = Batoms(label=obj.batoms.label)
        for i in v:
            setattr(batoms.bonds[i], self.key, getattr(self, self.key))
        # batoms.draw()
        context.view_layer.objects.active = obj
        bpy.ops.object.mode_set(mode="EDIT")
        return {"FINISHED"}


class BondOrderAutoSet(OperatorBatoms):
    bl_idname = "bond.bond_order_auto_set"
    bl_label = "Set bond order"
    bl_description = "Set bond order to a Batoms"

    def execute(self, context):
        obj = context.object
        batoms = Batoms(label=context.object.batoms.label)
        batoms.bond.bond_order_auto_set()
        context.view_layer.objects.active = obj
        return {"FINISHED"}


class BondShowHydrogenBond(OperatorBatoms):
    bl_idname = "bond.show_hydrogen_bond"
    bl_label = "Show hydrogen bond"
    bl_description = "Show hydrogen bond."

    def execute(self, context):
        obj = context.object
        batoms = Batoms(label=context.object.batoms.label)
        batoms.bond.show_hydrogen_bond = not batoms.bond.show_hydrogen_bond
        context.view_layer.objects.active = obj
        return {"FINISHED"}


class BondShowSearch(OperatorBatoms):
    bl_idname = "bond.show_search"
    bl_label = "Show atoms by searching bonds"
    bl_description = "Show atoms by searching bonds."

    def execute(self, context):
        batoms = Batoms(label=context.object.batoms.label)
        batoms.bond.show_search = not batoms.bond.show_search
        batoms.bond.update()
        context.view_layer.objects.active = batoms.obj
        return {"FINISHED"}



import bpy
from batoms.batoms import Batoms
from ase.build import molecule

try:
    from _common_helpers import has_display, set_cycles_res

    use_cycles = not has_display()
except ImportError:
    use_cycles = False

extras = dict(engine="cycles") if use_cycles else {}


def test_polyhedra_species(tio2):
    """
    This is an example to show different polyhedra for the same element, but different species.
    """
    assert len(tio2.polyhedra.settings) == 1
    # We set the first one to a new species Ti_1.
    tio2.replace([0], "Ti_1")
    tio2.bond.settings
    tio2.model_style = 2
    assert len(tio2.polyhedra) == 12
    # remove Ti_1 to polyhedra.
    tio2.polyhedra.settings.remove("Ti_1")
    assert len(tio2.polyhedra.settings) == 1
    tio2.model_style = 2
    assert len(tio2.polyhedra) == 6


def test_polyhedra_molecule():
    bpy.ops.batoms.delete()
    ch4 = molecule("CH4")
    mh4 = ch4.copy()
    mh4.translate([2, 2, 0])
    mh4[0].symbol = "N"
    ch4 = ch4 + mh4
    ch4 = Batoms("ch4", from_ase=ch4)
    ch4.model_style = 2
    ch4.bond.settings[("C", "H")].polyhedra = True
    ch4.model_style = 1


def test_polyhedra_crystal(tio2):
    tio2.model_style = 2
    tio2 = tio2 * [3, 3, 3]
    tio2.pbc = False
    tio2.model_style = 2


def test_polyhedra_setting(ch4):
    ch4.bond.settings[("C", "H")].polyhedra = True
    ch4.model_style = 2
    ch4.pbc = True
    ch4.cell = [3, 3, 3]
    ch4 = ch4 * [2, 1, 1]
    sel1 = ch4.selects.add("sel1", [0, 1, 2, 3, 4])
    sel1.model_style = 1
    ch4.polyhedra.settings.remove("C")
    assert len(ch4.polyhedra.settings) == 1
    ch4.polyhedra.settings.add("C")
    assert len(ch4.polyhedra.settings) == 2
    ch4.polyhedra.settings["C"] = {"color": [0.8, 0.1, 0.3, 0.3]}
    ch4.model_style = 2
    ch4.render.engine = "workbench"
    if use_cycles:
        set_cycles_res(ch4)
    ch4.get_image([1, 1, 0], output="polyhedra.png", **extras)



try:
    from _common_helpers import has_display

    use_cycles = not has_display()
except ImportError:
    use_cycles = False

extras = dict(engine="cycles") if use_cycles else {}


def test_highlight(ch4):
    """ """
    ch4.selects.add("s1", [2])
    ch4.highlight.settings["s1"] = {
        "select": "s1",
        "scale": 0.5,
        "color": (0.5, 0.5, 0, 0.4),
    }
    ch4.highlight.draw()
    assert len(ch4.highlight.obj.data.vertices) == 1
    ch4.get_image(output="highlight_ch4.png")
    # area = ch4.molecular_surface.get_psasa()


def test_highlight_style(ch4):
    """ """
    ch4.selects.add("s1", [2])
    ch4.highlight.settings["s1"] = {
        "select": "s1",
        "scale": 0.8,
        "color": (0.5, 0.5, 0, 0.4),
        "style": "1",
    }
    ch4.highlight.draw()
    assert len(ch4.highlight.obj.data.vertices) == 1
    ch4.get_image(output="highlight_ch4_cube.png")
    # area = ch4.molecular_surface.get_psasa()


def test_highlight_multi(ch4):
    """ """
    ch4.selects.add("s1", [2])
    ch4.selects.add("s2", [3])
    ch4.highlight.settings["s1"] = {
        "select": "s1",
        "scale": 0.5,
        "color": (0.5, 0.5, 0, 0.4),
    }
    ch4.highlight.settings["s2"] = {
        "select": "s2",
        "scale": 0.8,
        "color": (0, 0.5, 0.5, 0.4),
    }
    ch4.highlight.draw()
    assert len(ch4.highlight.obj.data.vertices) == 2
    ch4.get_image(output="highlight_ch4_multi.png")
    # area = ch4.molecular_surface.get_psasa()



import bpy
from bpy.props import (
    StringProperty,
    BoolProperty,
    IntProperty,
    FloatProperty,
    EnumProperty,
    FloatVectorProperty,
    CollectionProperty,
)

from batoms.internal_data import Base


class HighlightSetting(Base):
    """ """

    flag: BoolProperty(name="flag", default=False)
    name: StringProperty(name="name")
    label: StringProperty(name="label", default="")
    scale: FloatProperty(name="scale", default=1.0)
    select: StringProperty(name="select", default="all")
    color: FloatVectorProperty(
        name="color",
        size=4,
        subtype="COLOR",
        min=0,
        max=1,
        default=[1, 1, 0, 0.4],
    )
    style: EnumProperty(
        name="style",
        description="bond style",
        items=(("0", "Sphere", ""), ("1", "Cubic", "")),
        default="0",
    )

    def as_dict(self) -> dict:
        setdict = {
            "flag": self.flag,
            "scale": self.scale,
            "name": self.name,
            "select": self.select,
            "material_style": self.material_style,
            "color": self.color[:],
            "style": self.style,
        }
        return setdict

    def __repr__(self) -> str:
        s = "-" * 60 + "\n"
        s = "Name      select    scale        color            \n"
        s += "{:10s}   {:1.2f} {:1.2f} [{:1.2f}  {:1.2f}  {:1.2f}   {:1.2f}] \n".format(
            self.name,
            self.select,
            self.scale,
            self.color[0],
            self.color[1],
            self.color[2],
            self.color[3],
        )
        s += "-" * 60 + "\n"
        return s


class Highlight(bpy.types.PropertyGroup):
    """This module defines the highlight properties to extend
    Blender’s internal data.

    """

    active: BoolProperty(name="active", default=False)
    show: BoolProperty(name="show", default=True)
    settings: CollectionProperty(name="HighlightSetting", type=HighlightSetting)

    ui_list_index: IntProperty(name="ui_list_index", default=0)

    def as_dict(self) -> dict:
        setdict = {}
        return setdict



import bpy
from batoms.pdbparser import read_pdb
from batoms.batoms import Batoms


def test_ribbon():
    bpy.ops.batoms.delete()
    atoms = read_pdb("../tests/datas/1ema.pdb")  # 1ema, 1tim, 4hhb
    protein = Batoms("protein", from_ase=atoms)
    protein.ribbon.draw()
    sel1 = protein.selects.add("sel1", "sheet A-160-A-170")
    sel1.show = True
    sel1.model_style = 1
    bpy.ops.batoms.delete()



import bpy
import importlib


"""
import pathlib
import os
from time import time
cwd = os.path.dirname(pathlib.Path(__file__))
plugin_folders = []

tstart = time()
for f in os.listdir(cwd):
     if os.path.isdir(os.path.join(cwd, f)) and f != '__pycache__':
         plugin_folders.append(f)

plugin_info = {}
for name in plugin_folders:
    info = getattr(importlib.import_module("batoms.plugins.{}".format(name)), 'plugin_info')
    plugin_info[name] = info
print("Read plugin info: {:1.2f}".format(time() - tstart))
"""


plugin_info = {
    "highlight": ["highlight", "Highlight", "Bhighlight"],
    "cavity": ["cavity", "Cavity", "Bcavity"],
    "crystal_shape": ["crystal_shape", "CrystalShape", "Bcrystalshape"],
    "lattice_plane": ["lattice_plane", "LatticePlane", "Blatticeplane"],
    "isosurface": ["isosurface", "Isosurface", "Bisosurface"],
    "molecular_surface": ["molecular_surface", "MolecularSurface", "Bmolecularsurface"],
    "magres": ["magres", "Magres", "Bmagres"],
    "template": ["template", "Template", "Btemplate"],
}


def enable_plugin():
    # in case of using Blender as module
    # the "batoms" addon is added, but doesn't show up in the preferences
    if "batoms" not in bpy.context.preferences.addons:
        return
    for key in plugin_info.keys():
        if getattr(bpy.context.preferences.addons["batoms"].preferences, key):
            plugin = importlib.import_module("batoms.plugins.{}".format(key))
            plugin.register_class()


def disable_plugin():
    if "batoms" not in bpy.context.preferences.addons:
        return
    for key in plugin_info.keys():
        if getattr(bpy.context.preferences.addons["batoms"].preferences, key):
            plugin = importlib.import_module("batoms.plugins.{}".format(key))
            plugin.unregister_class()



"""Definition of the isosurface class.

This module defines the isosurface object in the Batoms package.

"""

import bpy
import numpy as np
from batoms.base.object import BaseObject
from .setting import IsosurfaceSettings
import logging

logger = logging.getLogger(__name__)


class Isosurface(BaseObject):
    def __init__(
        self,
        label=None,
        location=np.array([0, 0, 0]),
        batoms=None,
    ):
        """Isosurface Class

        Args:
            label (_type_, optional): _description_. Defaults to None.
            location (_type_, optional): _description_. Defaults to np.array([0, 0, 0]).
            batoms (_type_, optional): _description_. Defaults to None.
        """
        #
        self.batoms = batoms
        self.label = label
        name = "isosurface"
        BaseObject.__init__(self, label, name)
        self.settings = IsosurfaceSettings(self.label, batoms=batoms, parent=self)
        self.settings.bpy_data.active = True

    def build_isosurface(self, cell):
        isosurface = {}
        for iso in self.settings.bpy_setting:
            if iso.volumetric_data == "":
                volume = self.batoms.volumetric_data[
                    self.batoms.volumetric_data.bpy_setting[0].name
                ]
            else:
                volume = self.batoms.volumetric_data[iso.volumetric_data]
            scaled_verts, faces = calc_isosurface(volume, cell, iso.level)
            # color by another volumetric data
            if iso.color_by != "None":
                from batoms.utils import map_volumetric_data

                data = map_volumetric_data(
                    self.batoms.volumetric_data[iso.color_by], scaled_verts
                )
                attribute_data = (data - np.min(data)) / (np.max(data) - np.min(data))
                color_by_attribute = {
                    "attribute_name": "{}_data".format(iso.color_by),
                    "ValToRGB": [iso.color1[:], iso.color2[:]],
                }
            else:
                color_by_attribute = None
                attribute_data = None

            # back to cell
            verts = scaled_verts.dot(cell)
            # verts -= self.batoms.cell.origin
            isosurface[iso.name] = {
                "vertices": verts,
                "edges": [],
                "faces": faces,
                "material_style": iso.material_style,
                "color": iso.color,
                "color_by": iso.color_by,
                "color_by_attribute": color_by_attribute,
                "battr_inputs": {"Bisosurface": iso.as_dict()},
                "attribute_data": attribute_data,
            }
        return isosurface

    def build_materials(
        self,
        name,
        color,
        node_inputs=None,
        material_style="default",
        color_by_attribute=None,
    ):
        """ """
        from batoms.material import create_material

        if name in bpy.data.materials:
            mat = bpy.data.materials.get(name)
            bpy.data.materials.remove(mat, do_unlink=True)
        mat = create_material(
            name,
            color=color,
            node_inputs=node_inputs,
            material_style=material_style,
            backface_culling=False,
            color_by_attribute=color_by_attribute,
        )
        return mat

    def draw(self, isosurface_name="ALL"):
        """Draw isosurface."""
        from batoms.draw import draw_surface_from_vertices
        from batoms.utils.butils import clean_coll_object_by_type

        # delete old isosurface
        clean_coll_object_by_type(self.batoms.coll, "ISOSURFACE")
        isosurface = self.build_isosurface(self.batoms.cell)
        for name, isosurface_data in isosurface.items():
            if isosurface_name.upper() != "ALL" and name.name != isosurface_name:
                continue
            name = "%s_%s_%s" % (self.label, "isosurface", name)
            self.delete_obj(name)
            obj = draw_surface_from_vertices(
                name,
                datas=isosurface_data,
                coll=self.batoms.coll,
            )
            obj.batoms.type = "ISOSURFACE"
            obj.batoms.label = self.label
            obj.parent = self.batoms.obj
            if isosurface_data["attribute_data"] is not None:
                from batoms.utils.attribute import set_mesh_attribute

                obj.data.attributes.new(
                    name=isosurface_data["color_by_attribute"]["attribute_name"],
                    type="FLOAT",
                    domain="POINT",
                )
                set_mesh_attribute(
                    obj,
                    isosurface_data["color_by_attribute"]["attribute_name"],
                    isosurface_data["attribute_data"],
                )
            # material
            mat = self.build_materials(
                name,
                isosurface_data["color"],
                material_style=isosurface_data["material_style"],
                color_by_attribute=isosurface_data["color_by_attribute"],
            )
            obj.data.materials.append(mat)

    @property
    def setting(self):
        from batoms.utils import deprecated

        """setting object."""
        deprecated(
            '"setting" will be deprecated in the furture, please use "settings".'
        )
        return self.settings

    def as_dict(self):
        """ """
        data = {}
        data["settings"] = self.settings.as_dict()
        data.update(self.settings.bpy_data.as_dict())
        return data


def calc_isosurface(
    volumetric_data, cell, level, gradient_direction="descent", step_size=1
):
    """

    Computes an isosurface from a volumetric_data grid.

    Parameters:

    volumetric_data: np.array

    cell: np.array

    level: float

    """
    from skimage import measure

    spacing = tuple(1.0 / np.array(volumetric_data.shape))
    mlevel = np.mean(volumetric_data)
    if not level:
        level = mlevel * 10
    # print('iso level: {0:1.9f}, iso mean: {1:1.9f}'.format(level, mlevel))
    scaled_verts, faces, normals, values = measure.marching_cubes(
        volumetric_data,
        level=level,
        spacing=spacing,
        gradient_direction=gradient_direction,
        allow_degenerate=False,
        step_size=step_size,
    )
    faces = list(faces)
    return scaled_verts, faces



"""Definition of the parent plugin class.

"""

import bpy
from time import time
from batoms.base.collection import Setting
from batoms.base.object import BaseObject
import logging

logger = logging.getLogger(__name__)


class PluginSettings(Setting):
    def __init__(self, label, parent) -> None:
        """Plugin Settings
        This object store the plugin setting.

        Parameters:

        label: str
            The label define the batoms object that
            a Setting belong to.
        parent: Plugin Class

        """
        Setting.__init__(self, label, coll_name=label)
        self.label = label
        self.parent = parent
        self.name = "B{}".format(parent.name)


class PluginObject(BaseObject):
    def __init__(
        self,
        label,
        name="plugin",
        batoms=None,
    ):
        """Plugin Class

        Args:
            label (string): _description_.
            batoms (Batoms, optional): _description_. Defaults to None.
        """
        #
        self.batoms = batoms
        self.label = label
        self.name = name
        obj_name = "%s_%s" % (label, name)
        BaseObject.__init__(self, obj_name)
        self.settings = PluginSettings(self.label, parent=self)

    def build_materials(
        self,
        name,
        color,
        node_inputs=None,
        material_style="default",
        backface_culling=False,
        vertex_color=None,
    ):
        """Create materials

        Args:
            name (string): _description_
            color (array, (1x4)): _description_
            node_inputs (dict, optional): _description_. Defaults to None.
            material_style (string, optional): _description_. Defaults to 'default'.

        Returns:
            bpy.type.material: _description_
        """
        from batoms.material import create_material

        if name in bpy.data.materials:
            mat = bpy.data.materials.get(name)
            bpy.data.materials.remove(mat, do_unlink=True)
        mat = create_material(
            name,
            color=color,
            node_inputs=node_inputs,
            material_style=material_style,
            backface_culling=backface_culling,
            vertex_color=vertex_color,
        )
        return mat

    def draw(self):
        from batoms.utils.butils import clean_coll_object_by_type

        # delete old object
        clean_coll_object_by_type(self.batoms.coll, "PLUGIN")
        for setting in self.settings.bpy_setting:
            self.draw_item(setting)

    def draw_item(self, setting):
        """ """
        tstart = time()
        indices = self.batoms.selects[setting.select].indices
        if len(indices) == 0:
            return
        # select atoms
        self.batoms.positions[indices]
        logger.debug("Draw Template: %s" % (time() - tstart))

    @property
    def objs(self):
        objs = {}
        for setting in self.settings.bpy_setting:
            name = "{}_{}_{}".format(self.label, self.name, setting.name)
            obj = bpy.data.objects.get(name)
            objs[name] = obj
        return objs

    @property
    def mat(self):
        return bpy.data.materials.get(self.name)

    def as_dict(self):
        """ """
        data = {}
        data["settings"] = self.settings.as_dict()
        data.update(self.settings.bpy_data.as_dict())
        return data

    @property
    def show(self):
        return self.get_show()

    @show.setter
    def show(self, state):
        self.set_show(state)

    def get_show(self):
        return self.settings.bpy_data.show

    def set_show(self, state):
        self.settings.bpy_data.show = state
        for name, obj in self.objs.items():
            obj.hide_render = not state
            obj.hide_set(not state)



import numpy as np
from math import pi


def ellipse(n, w, h):
    profile = np.zeros((n, 3))
    t = np.linspace(0, 1, n, endpoint=False)
    a = t * 2 * pi + pi / 4
    profile[:, 0] = np.cos(a) * w
    profile[:, 2] = np.sin(a) * h
    return profile


def rectangle(w, h):
    profile = np.array(
        [
            [-1 * w, 0, -1 * h],
            [-1 * w, 0, 1 * h],
            [1 * w, 0, 1 * h],
            [1 * w, 0, -1 * h],
        ]
    )

    return profile


def build_mesh(vertices, normals, sides, profile, scales=None):
    """
    todo: improve performance
    """
    # calc side interp
    # tstart = time()
    n = len(vertices)
    m = len(profile)
    forwards = np.cross(normals, sides)
    rotations = np.zeros((n, 3, 3))
    sides = sides.reshape(n, 1, 3)
    forwards = forwards.reshape(n, 1, 3)
    normals = normals.reshape(n, 1, 3)
    # build new axis for a given vertices
    newAxis = np.concatenate((sides, forwards, normals), axis=1)
    # rotate and scale the profile
    rotations = np.linalg.inv(newAxis)
    if scales is not None:
        rotations = rotations * scales[:, None]
    vertices = np.tile(vertices, (m, 1))
    #
    for i in range(m):
        vertices[i * n : (i + 1) * n] += np.dot(rotations, profile[i])
    #
    faces = np.zeros((m * (n - 1), 4), dtype=int)
    for i in range(m - 1):
        faces[i * (n - 1) : (i + 1) * (n - 1), 0] = np.arange(i * n, (i + 1) * n - 1)
        faces[i * (n - 1) : (i + 1) * (n - 1), 1] = np.arange(i * n + 1, (i + 1) * n)
        faces[i * (n - 1) : (i + 1) * (n - 1), 2] = np.arange(
            (i + 1) * n + 1, (i + 2) * n
        )
        faces[i * (n - 1) : (i + 1) * (n - 1), 3] = np.arange(
            (i + 1) * n, (i + 2) * n - 1
        )
    i = m - 1
    faces[i * (n - 1) : (i + 1) * (n - 1), 0] = np.arange(i * n, (i + 1) * n - 1)
    faces[i * (n - 1) : (i + 1) * (n - 1), 1] = np.arange(i * n + 1, (i + 1) * n)
    faces[i * (n - 1) : (i + 1) * (n - 1), 2] = np.arange(1, n)
    faces[i * (n - 1) : (i + 1) * (n - 1), 3] = np.arange(0, n - 1)
    # cap
    faces = faces.tolist()
    faces.append([i * n for i in range(m)])
    faces.append([(i + 1) * n - 1 for i in range(m)])
    # print('build mesh: %s'%(time() - tstart))
    return vertices, faces


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    # profile = ellipse(16, 1, 1)
    profile = rectangle(1, 1)
    plt.plot(profile[:, 0], profile[:, 1])
    plt.show()



import bpy
from batoms import Batoms
import pytest


def test_pybel_smiles():
    pytest.importorskip("openbabel")
    from openbabel import pybel

    bpy.ops.batoms.delete()
    smiles = "CCO"
    mol = pybel.readstring("smi", smiles)
    mol.make3D(forcefield="mmff94", steps=100)
    mol = Batoms("c2H6O", from_pybel=mol)
    assert len(mol) == 9


def test_pybel_bond_order():
    pytest.importorskip("openbabel")
    bpy.ops.batoms.delete()
    bpy.ops.batoms.molecule_add(label="c6h6", formula="C6H6")
    c6h6 = Batoms("c6h6")
    c6h6.model_style = 1
    c6h6.bond.bond_order_auto_set()
    assert c6h6.bond.arrays["bond_order"][0] == 2



import bpy
import numpy as np
from batoms.utils.butils import object_mode

# from time import time


class BaseCollection:
    def __init__(self, coll_name):
        self.coll_name = coll_name

    def set_collection(self, label):
        """
        build main collection and its child collections.
        """
        if bpy.data.collections.get(label):
            raise Exception("Failed, the name %s already in use!" % label)
        coll = bpy.data.collections.new(label)
        return coll

    @property
    def coll(self):
        return self.get_coll()

    def get_coll(self):
        coll = bpy.data.collections.get(self.coll_name)
        if coll is None:
            raise ("No collection: %s" % self.coll_name)
        return coll

    @property
    def scene(self):
        return self.get_scene()

    def get_scene(self):
        return bpy.context.scene

    def translate(self, displacement):
        """Translate atomic positions.

        The displacement argument is an xyz vector.

        For example, move H2o molecule by a vector [0, 0, 5]

        >>> h2o.translate([0, 0, 5])
        """
        object_mode()
        bpy.ops.object.select_all(action="DESELECT")
        self.obj.select_set(True)
        bpy.ops.transform.translate(value=displacement)

    def rotate(self, angle, axis="Z", orient_type="GLOBAL"):
        """Rotate atomic based on a axis and an angle.

        Parameters:

        angle: float
            Angle that the atoms is rotated around the axis.
        axis: str
            'X', 'Y' or 'Z'.

        For example, rotate h2o molecule 90 degree around 'Z' axis:

        >>> h2o.rotate(90, 'Z')

        """
        object_mode()
        bpy.ops.object.select_all(action="DESELECT")
        self.obj.select_set(True)
        angle = angle / 180.0 * np.pi
        bpy.ops.transform.rotate(
            value=angle, orient_axis=axis.upper(), orient_type=orient_type
        )

    def mirror(self, axis="Z", orient_type="GLOBAL"):
        """mirror atomic based on a axis.

        Parameters:

        angle: float
            Angle that the atoms is mirrord around the axis.
        axis: str
            Constraint Axis: 'X', 'Y' or 'Z'.

        For example, mirror h2o using 'YZ' plane:

        >>> h2o.mirror('X')

        """
        object_mode()
        bpy.ops.object.select_all(action="DESELECT")
        self.obj.select_set(True)
        constraint_axis = [False, False, False]
        constraint_axis["XYZ".index(axis.upper())] = True
        bpy.ops.transform.mirror(
            constraint_axis=constraint_axis, orient_type=orient_type
        )


def tuple2string(index):
    if isinstance(index, (str, int, float)):
        return str(index)
    name = str(index[0])
    for key in index[1:]:
        name += "-%s" % key
    return name


class Setting:
    """
    Setting object

    The Setting object store the additional information for batoms.
    The inforation is a colleciton of data. For example, species, bond pairs.
    In Batoms, collection properties are added to a bpy.type.collection
    or a bpy.type.object.

    Setting has "add", "remove", "find" and "extend" operators.

    Note: In blender, the colleciton properties is a list.
    Here, we build our Setting as a dict.

    Parameters:

    label: str
        The label define the batoms object that a Setting belong to.

    """

    def __init__(self, label, coll_name=None, obj_name=None) -> None:
        self.label = label
        self.name = "base"
        self.coll_name = coll_name
        self.obj_name = obj_name

    @property
    def coll(self):
        """collection

        Returns:
            bpy.type.collection: collection which setting atached to.
        """
        return self.get_coll()

    def get_coll(self):
        coll = bpy.data.collections.get(self.coll_name)
        if coll is None:
            coll = bpy.data.collections.new(self.coll_name)
            bpy.context.scene.collection.children.link(coll)
        return coll

    @property
    def obj(self):
        """object

        Returns:
            bpy.type.object: object which setting atached to.
        """
        return self.get_obj()

    def get_obj(self):
        obj = bpy.data.objects.get(self.obj_name)
        if obj is None:
            raise KeyError("%s object is not exist." % self.obj_name)
        return obj

    @property
    def bpy_setting(self):
        """bpy_setting properties

        Returns:
            bpy.props.bpy_settingProperty: colleciton of the properties
        """
        return self.get_bpy_setting()

    @property
    def bpy_data(self):
        """Get internal data

        Internal data are data attached to a Blender object or collection.
        Raises:
            KeyError: _description_

        Returns:
            _type_: _description_
        """
        if self.coll_name:
            coll = bpy.data.collections.get(self.coll_name)
            data = getattr(coll, self.name)
        elif self.obj_name:
            obj = bpy.data.objects.get(self.obj_name)
            data = getattr(obj, self.name)
        else:
            raise KeyError("The collection property {} not exist!".format(self.name))
        return data

    @property
    def ui_list_index(self):
        return self.get_ui_list_index()

    def get_ui_list_index(self):
        return self.bpy_data.ui_list_index

    @ui_list_index.setter
    def ui_list_index(self, value):
        self.set_ui_list_index(value)

    def set_ui_list_index(self, value):
        self.bpy_data.ui_list_index = value

    def get_bpy_setting(self):
        return self.bpy_data.settings

    def keys(self):
        return self.as_dict().keys()

    def values(self):
        return self.as_dict().values()

    def items(self):
        return self.as_dict().items()

    def as_dict(self):
        """Return a dict of all data in the collection.

        Returns:
            dict: dict of properties.
        """
        data = {}
        for b in self.bpy_setting:
            data[b.name] = b.as_dict()
        return data

    def __getitem__(self, key):
        item = self.find(key)
        if item is None:
            raise Exception("%s not in %s setting" % (key, self.name))
        return item

    def __setitem__(self, key, setdict):
        """
        Set properties
        """
        subset = self.find(key)
        if subset is None:
            subset = self.bpy_setting.add()
            name = tuple2string(key)
            subset.name = name
            self.ui_list_index = len(self) - 1
        subset.label = self.label
        subset.flag = True
        for key, value in setdict.items():
            setattr(subset, key, value)

    def __delitem__(self, key):
        self.remove(key)

    def from_dict(self, datas):
        if isinstance(datas, dict):
            datas = [datas]
        for data in datas:
            subset = self.find(data["name"])
            if subset is None:
                subset = self.bpy_setting.add()
            for key, value in data.items():
                setattr(subset, key, value)
            subset.label = self.label
            subset.flag = True

    def copy(self, label):
        object_mode()
        setting = self.__class__(label)
        for key, b in self.data.items():
            setting[key] = b.as_dict()
        return setting

    def add(self, key):
        """Add default item

        Args:
            key (tuple, list): _description_
        """
        self[key] = {}

    def remove(self, key):
        """Remove item

        Args:
            key (tuple, list): _description_

        Raises:
            Exception: _description_
        """
        item = self.find(key)
        if item is not None:
            i = self.bpy_setting.find(item.name)
            self.bpy_setting.remove(i)
            self.ui_list_index = min(max(0, self.ui_list_index - 1), len(self) - 1)
        else:
            raise Exception("%s is not in %s" % (key, self.name))

    def __add__(self, other):
        self += other
        return self

    def __iadd__(self, other):
        self.extend(other)
        return self

    def extend(self, other):
        for key, value in other.data.items():
            self[key] = value.as_dict()
        return self

    def __iter__(self):
        item = self.bpy_setting
        for i in range(len(item)):
            yield item[i]

    def __len__(self):
        return len(self.bpy_setting)

    def find(self, name):
        """Wrapper function for self.bpy_setting.find().

        Note: In blender, the colleciton properties is a list.
        self.bpy_setting.find will return the index of a item.
        Here, we build our Setting as a dict.
        this find will return the item itself.

        Args:
            name (str, tuple, list): _description_

        Returns:
            Setting: setting object
        """
        if isinstance(name, (tuple, list)):
            name = tuple2string(name)
        #
        i = self.bpy_setting.find(str(name))
        if i == -1:
            # print('%s is not in %s'%(name, self.name))
            return None
        else:
            return self.bpy_setting[i]

    def __repr__(self) -> str:
        s = "-" * 60 + "\n"
        s = "Name\n"
        for b in self.bpy_setting:
            s += "{0:10s}  \n".format(b.name)
        s += "-" * 60 + "\n"
        return s

    def delete_obj(self, name):
        if name in bpy.data.objects:
            obj = bpy.data.objects.get(name)
            bpy.data.objects.remove(obj, do_unlink=True)



import bpy
import logging
import importlib
import pkgutil


package = "batoms"
addon = bpy.context.preferences.addons[package]
preferences = addon.preferences


def test_enable_disable_plugin():
    from bpy.types import Collection

    assert preferences.magres is True
    assert hasattr(Collection, "Bmagres")
    preferences.magres = False
    assert not hasattr(Collection, "Bmagres")
    preferences.magres = True
    assert hasattr(Collection, "Bmagres")


def create_molecule():
    bpy.ops.batoms.delete()
    from batoms import Batoms
    from ase.build import molecule

    Batoms(from_ase=molecule("CO"))
    bpy.ops.batoms.delete()
    return


def read_log_lines(logger):
    """Ugly patch for reading the logger contents,
    since pytest caplog not working properly in blender (yet)
    """
    fh = logger.handlers[0]
    filename = fh.baseFilename
    with open(filename, "r") as fd:
        lines = fd.readlines()
    return lines


def import_submodules(package, recursive=True):
    """Import all submodules of a module, recursively, including subpackages
    https://stackoverflow.com/questions/3365740/how-to-import-all-submodules
    :param package: package (name or actual module)
    :type package: str | module
    :rtype: dict[str, types.ModuleType]
    """
    if isinstance(package, str):
        package = importlib.import_module(package)
    results = {}
    for loader, name, is_pkg in pkgutil.walk_packages(package.__path__):
        full_name = package.__name__ + "." + name
        results[full_name] = importlib.import_module(full_name)
        if recursive and is_pkg:
            results.update(import_submodules(full_name))
    return results


def test_child_loggers():
    """Make sure all submodules of batoms follow the setLevel rule
    https://stackoverflow.com/questions/3365740/how-to-import-all-submodules
    """
    all_submodules = import_submodules("batoms")
    preferences.logging_level = "WARNING"
    for name, mod in all_submodules.items():
        try:
            logger = mod.logger
            print(f"Test {name}.logger")
            assert logger.name == name
            assert logger.getEffectiveLevel() == logging.WARNING
        except AttributeError:
            print(f"{name}.logger not found. skip")
            continue


def test_logging_level():
    """logging level.
    Since now all the child loggers are "delegation to the parent"
    https://docs.python.org/3/library/logging.html#logging.Logger.setLevel,
    explicitly testing logger.level will not work. Instead, use getEffectiveLevel to check
    the effect of propagation
    """
    from batoms.batoms import logger

    for level in ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]:
        preferences.logging_level = level
        assert logger.getEffectiveLevel() == getattr(logging, level)
    # bpy.context.preferences.addons[package].preferences.logging_level = "INFO"
    # assert(logger.level == 20)
    # bpy.context.preferences.addons[package].preferences.logging_level = "WARNING"
    # assert(logger.level == 30)


def test_logging_level_emit():
    """Test if setting logging level hierachically works
    Set the logging level to INFO, adding Batoms shows the timing info
    Set the level to WARNING and higher, timing info is supressed
    """
    root_logger = logging.getLogger("batoms")
    # INFO level reveals the timing info
    preferences.logging_level = "INFO"
    create_molecule()
    lines1 = read_log_lines(root_logger)
    assert any(["Add object" in line for line in lines1])
    # WARNING level suppresses timing info
    preferences.logging_level = "WARNING"
    create_molecule()
    lines2 = read_log_lines(root_logger)[len(lines1) :]
    assert all(["Add object" not in line for line in lines2])



import bpy
from bpy.props import (
    StringProperty,
    BoolProperty,
    IntProperty,
    FloatProperty,
    EnumProperty,
    FloatVectorProperty,
    CollectionProperty,
)

from batoms.internal_data import Base


class CavitySetting(Base):
    """ """

    flag: BoolProperty(name="flag", default=False)
    name: StringProperty(name="name")
    label: StringProperty(name="label", default="")
    min: FloatProperty(name="min", default=5)
    max: FloatProperty(name="max", default=6)
    scale: FloatProperty(name="scale", default=1.0)
    resolution: FloatProperty(name="resolution", soft_min=0.5, soft_max=2, default=1.0)
    color: FloatVectorProperty(
        name="color",
        size=4,
        subtype="COLOR",
        min=0,
        max=1,
        default=[1, 1, 0, 1.0],
    )

    def as_dict(self) -> dict:
        setdict = {
            "flag": self.flag,
            "min": self.min,
            "max": self.max,
            "scale": self.scale,
            "name": self.name,
            "material_style": self.material_style,
            "color": self.color[:],
        }
        return setdict

    def __repr__(self) -> str:
        s = "-" * 60 + "\n"
        s = "Name      min    max        color            \n"
        s += "{:10s}   {:1.2f} {:1.2f} [{:1.2f}  {:1.2f}  {:1.2f}   {:1.2f}] \n".format(
            self.name,
            self.min,
            self.max,
            self.color[0],
            self.color[1],
            self.color[2],
            self.color[3],
        )
        s += "-" * 60 + "\n"
        return s


class Cavity(bpy.types.PropertyGroup):
    """This module defines the cavity properties to extend
    Blender’s internal data.

    """

    active: BoolProperty(name="active", default=False)
    model_style: EnumProperty(
        name="model_style",
        description="Structural models",
        items=(
            ("0", "Surface", "Surface"),
            ("1", "Dot", "Dot surface"),
            ("2", "Wireframe", "Use wireframe"),
        ),
        default="0",
    )
    show: BoolProperty(name="show", default=True)
    atomRadius: FloatProperty(name="atomRadius", default=0.5)
    minCave: FloatProperty(name="minCave", default=3)
    resolution: FloatProperty(name="resolution", default=1)
    settings: CollectionProperty(name="CavitySetting", type=CavitySetting)

    ui_list_index: IntProperty(name="ui_list_index", default=0)

    def as_dict(self) -> dict:
        setdict = {}
        return setdict



"""

Wrapper functions for ASE build

"""

from ase import build
from batoms.batoms import Batoms


def molecule(label, symbol, **kwargs):
    """_summary_

    Args:
        label (_type_): _description_
        symbol (_type_): _description_

    Returns:
        _type_: _description_
    """
    atoms = build.molecule(symbol, **kwargs)
    batoms = Batoms(label=label, from_ase=atoms)
    return batoms


def bulk(label, symbol, **kwargs):
    """_summary_

    Args:
        label (_type_): _description_
        symbol (_type_): _description_

    Returns:
        _type_: _description_
    """
    atoms = build.bulk(symbol, **kwargs)
    batoms = Batoms(label=label, from_ase=atoms)
    return batoms


def surface(label, lattice, indices, layers, **kwargs):
    """_summary_

    Args:
        label (_type_): _description_
        lattice (_type_): _description_
        indices (_type_): _description_
        layers (_type_): _description_

    Returns:
        _type_: _description_
    """
    if isinstance(lattice, Batoms):
        lattice = lattice.atoms
    atoms = build.surface(lattice, indices, layers, **kwargs)
    atoms.info.pop("species", None)
    batoms = Batoms(label=label, from_ase=atoms)
    return batoms



from .molecular_surface import MolecularSurface

from . import (
    bpy_data,
    ops,
    ui_list,
)

__all__ = ["MolecularSurface"]


classes = [
    # internal data first
    bpy_data.MolecularSurfaceSetting,
    bpy_data.MolecularSurface,
    ops.MolecularSurfaceAdd,
    ops.MolecularSurfaceRemove,
    ops.MolecularSurfaceDraw,
    ui_list.BATOMS_MT_molecular_surface_context_menu,
    ui_list.BATOMS_UL_molecular_surface,
    ui_list.BATOMS_PT_molecular_surface,
]


def register_class():
    from bpy.types import Collection, Object
    from bpy.props import PointerProperty
    from bpy.utils import register_class

    for cls in classes:
        register_class(cls)
    # attach to blender internal data
    Collection.Bmolecularsurface = PointerProperty(
        name="Bmolecularsurface", type=bpy_data.MolecularSurface
    )
    Object.Bmolecularsurface = PointerProperty(
        name="Bmolecularsurface", type=bpy_data.MolecularSurface
    )


def unregister_class():
    from bpy.types import Collection, Object
    from bpy.utils import unregister_class

    for cls in reversed(classes):
        unregister_class(cls)

    del Collection.Bmolecularsurface
    del Object.Bmolecularsurface



"""
Visualising tensors using ellipsoids.

1) magnetic shielding tensor
2) chemical shift tensor

"""
import bpy
from batoms.base.collection import Setting
import logging

logger = logging.getLogger(__name__)


class MagresSettings(Setting):
    def __init__(
        self,
        label,
        parent,
        scale=1.4,
        resolution=0.5,
    ) -> None:
        """Marges object
        The Marges object store the Marges information.

        Parameters:

        label: str
            The label define the batoms object that
            a Setting belong to.
        scale: float
            length of the largest tensor principal component [angstroms], default is 3.
        """
        Setting.__init__(self, label, coll_name=label)
        self.label = label
        self.name = "Bmagres"
        self.scale = scale
        self.parent = parent
        self.ms_name = "%s_ms" % self.label
        self.cs_name = "%s_cs" % self.label
        self.resolution = resolution
        if len(self) == 0:
            self["1"] = {"select": "all"}

    def add(self, name, datas={}):
        self[name] = datas

    def set_collection(self, label):
        """ """
        if not bpy.data.collections.get(label):
            coll = bpy.data.collections.new(label)
            self.parent.batoms.coll.children.link(coll)
            coll.batoms.type = "MAGRES"
            coll.batoms.label = label

    def __repr__(self) -> str:
        s = "-" * 60 + "\n"
        s = "name  type select  scale   resolution    color  \n"
        for magres in self.bpy_setting:
            s += "{:4s}   {:6s}   {:6s}   {:1.3f}  {:1.3f} ".format(
                magres.name, magres.type, magres.select, magres.scale, magres.resolution
            )
            s += "[{:1.1f}  {:1.1f}  {:1.1f}  {:1.1f}] \n".format(
                magres.color[0], magres.color[1], magres.color[2], magres.color[3]
            )
        s += "-" * 60 + "\n"
        return s



"""
Use ASE's build function
https://wiki.fysik.dtu.dk/ase/ase/build/build.html?highlight=nanotube#
"""
import bpy
from bpy.types import Operator
from bpy.props import (
    StringProperty,
    FloatProperty,
    FloatVectorProperty,
    BoolProperty,
)
from ase.build import molecule, bulk
from ase import Atoms
from batoms import Batoms


class AddMolecule(Operator):
    bl_idname = "batoms.molecule_add"
    bl_label = "Add molecule"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Add molecule"

    label: StringProperty(name="Label", default="", description="Label")

    formula: StringProperty(name="Formula", default="CH4", description="formula")

    def execute(self, context):
        if self.label == "":
            self.label = self.formula
        atoms = molecule(self.formula)
        if self.label in bpy.data.collections:
            self.label = "%s_001" % self.label
        batoms = Batoms(label=self.label, from_ase=atoms)
        batoms.model_style = 1
        batoms.obj.select_set(True)
        bpy.context.view_layer.objects.active = batoms.obj
        self.report({"INFO"}, "Add molecule {}".format(self.label))
        return {"FINISHED"}


class AddBulk(Operator):
    bl_idname = "batoms.bulk_add"
    bl_label = "Add bulk"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Add bulk"

    label: StringProperty(name="Label", default="", description="Label")

    formula: StringProperty(name="Formula", default="Au", description="formula")

    crystalstructure: StringProperty(
        name="Crystalstructure", default="", description="Crystalstructure"
    )

    latticeconstant: FloatVectorProperty(
        name="Lattice constant",
        default=(0, 0, 0),
        min=0,
        soft_max=100,
        description="Lattice constant",
    )

    alpha: FloatProperty(
        name="Angle",
        default=0,
        min=0,
        soft_max=360,
        description="Angle in degrees for rhombohedral lattice.",
    )

    covera: FloatProperty(
        name="Covera",
        default=1.6329931,
        min=0,
        soft_max=10,
        description="c/a ratio used for hcp. Default is ideal ratio: sqrt(8/3).",
    )

    orthorhombic: BoolProperty(
        name="Orthorhombic", default=False, description="Orthorhombic"
    )

    cubic: BoolProperty(name="Cubic", default=False, description="Cubic")

    def execute(self, context):
        if self.label == "":
            self.label = self.formula
        a = None if self.latticeconstant[0] == 0 else self.latticeconstant[0]
        b = None if self.latticeconstant[1] == 0 else self.latticeconstant[1]
        c = None if self.latticeconstant[2] == 0 else self.latticeconstant[2]
        crystalstructure = (
            None if self.crystalstructure == "" else self.crystalstructure
        )
        atoms = bulk(
            self.formula,
            crystalstructure=crystalstructure,
            a=a,
            b=b,
            c=c,
            orthorhombic=self.orthorhombic,
            cubic=self.cubic,
        )
        if self.label in bpy.data.collections:
            self.label = "%s_001" % self.label
        batoms = Batoms(label=self.label, from_ase=atoms)
        batoms.obj.select_set(True)
        bpy.context.view_layer.objects.active = batoms.obj
        self.report({"INFO"}, "Add bulk {}".format(self.label))
        return {"FINISHED"}


class AddAtoms(Operator):
    bl_idname = "batoms.atoms_add"
    bl_label = "Add atoms"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Add atoms"

    label: StringProperty(name="Label", default="", description="Label")

    formula: StringProperty(name="Formula", default="C", description="formula")

    def execute(self, context):
        if self.label == "":
            self.label = self.formula
        atoms = Atoms(self.formula)
        if self.label in bpy.data.collections:
            self.label = "%s_001" % self.label
        batoms = Batoms(label=self.label, from_ase=atoms)
        batoms.obj.select_set(True)
        bpy.context.view_layer.objects.active = batoms.obj
        self.report({"INFO"}, "Add atoms {}".format(self.label))
        return {"FINISHED"}


class AddSmiles(Operator):
    bl_idname = "batoms.smiles_add"
    bl_label = "Add Smiles"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Add Smiles"

    label: StringProperty(name="Label", default="", description="Label")

    smiles: StringProperty(name="smiles", default="CCO", description="smiles")

    def execute(self, context):
        from openbabel import pybel

        if self.label == "":
            self.label = self.smiles
        if self.label in bpy.data.collections:
            self.label = "%s_001" % self.label
        mol = pybel.readstring("smi", self.smiles)
        mol.make3D(forcefield="mmff94", steps=100)
        batoms = Batoms(label=self.label, from_pybel=mol)
        batoms.model_style = 1
        batoms.obj.select_set(True)
        bpy.context.view_layer.objects.active = batoms.obj
        self.report({"INFO"}, "Add smiles {}".format(self.label))
        return {"FINISHED"}



"""
"""

import bpy
from bpy.types import Menu, Panel, UIList


class BATOMS_MT_polyhedra_context_menu(Menu):
    bl_label = "polyhedra Specials"
    bl_idname = "BATOMS_MT_polyhedra_context_menu"

    def draw(self, _context):
        layout = self.layout
        op = layout.operator("batoms.polyhedra_add", icon="ADD", text="Add polyhedra")
        layout.separator()
        op = layout.operator(
            "batoms.polyhedra_remove", icon="X", text="Delete All polyhedra"
        )
        op.all = True


class BATOMS_UL_polyhedra(UIList):
    def draw_item(
        self, _context, layout, _data, item, icon, active_data, _active_propname, index
    ):
        polyhedra = item
        custom_icon = "OBJECT_DATAMODE"
        if self.layout_type in {"DEFAULT", "COMPACT"}:
            split = layout.split(factor=0.66, align=False)
            split.prop(polyhedra, "name", text="", emboss=False, icon=custom_icon)
            row = split.row(align=True)
            row.emboss = "NONE_OR_STATUS"
            # row.prop(polyhedra, "distance", text="")
        elif self.layout_type == "GRID":
            layout.alignment = "CENTER"
            layout.label(text="", icon=custom_icon)


class BATOMS_PT_polyhedra(Panel):
    bl_label = "Polyhedra"
    bl_category = "Batoms"
    bl_idname = "BATOMS_PT_polyhedra"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"DEFAULT_CLOSED"}

    COMPAT_ENGINES = {"BLENDER_RENDER", "BLENDER_EEVEE", "BLENDER_WORKBENCH"}

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj:
            return obj.batoms.type != "OTHER"
        else:
            return False

    def draw(self, context):
        layout = self.layout

        ob = context.object
        ba = bpy.data.collections[ob.batoms.label].Bpolyhedra
        if len(ba.settings) > 0:
            kb = ba.settings[ba.ui_list_index]
        else:
            kb = None

        row = layout.row()

        rows = 3
        if kb:
            rows = 5

        row.template_list(
            "BATOMS_UL_polyhedra", "", ba, "settings", ba, "ui_list_index", rows=rows
        )

        col = row.column(align=True)
        op = col.operator("batoms.polyhedra_add", icon="ADD", text="")
        op = col.operator("batoms.polyhedra_remove", icon="REMOVE", text="")
        if kb is not None:
            op.species = kb.species
        col.separator()

        col.menu("BATOMS_MT_polyhedra_context_menu", icon="DOWNARROW_HLT", text="")

        if kb:
            col.separator()

            sub = col.column(align=True)

            split = layout.split(factor=0.4)
            row = split.row()

            row = split.row()
            row.alignment = "RIGHT"

            sub = row.row(align=True)
            sub.label()  # XXX, for alignment only

            sub = row.row()
            layout.use_property_split = True
            row = layout.row()
            col = layout.column()
            sub = col.column(align=True)
            col.prop(kb, "material_style", text="material_style")
            sub.prop(kb, "color", text="Color")
            # col.prop(kb, "show_edge",  text="Show edge")
            # col.prop(kb, "width",  text="Width")
            op = layout.operator(
                "batoms.polyhedra_draw", icon="GREASEPENCIL", text="Update"
            )



import bmesh
from bpy.types import Operator
from bpy.props import BoolProperty, FloatProperty, StringProperty
from batoms import Batoms
from batoms.ops.base import OperatorBatoms


class PolyhedraAdd(OperatorBatoms):
    bl_idname = "batoms.polyhedra_add"
    bl_label = "Add polyhedra"
    bl_description = "Add polyhedra to a Batoms"

    species: StringProperty(
        name="species", default="", description="species to be added"
    )

    def execute(self, context):
        if self.species == "":
            return {"FINISHED"}
        obj = context.object
        batoms = Batoms(label=context.object.batoms.label)
        batoms.polyhedra.settings.add(self.species)
        batoms.coll.Bpolyhedra.ui_list_index = len(batoms.polyhedra.settings) - 1
        context.view_layer.objects.active = obj
        return {"FINISHED"}


class PolyhedraRemove(OperatorBatoms):
    bl_idname = "batoms.polyhedra_remove"
    bl_label = "Remove polyhedra"
    bl_description = "Remove polyhedra to a Batoms"

    species: StringProperty(
        name="species", default="C", description="species to be removed"
    )

    all: BoolProperty(name="all", default=False, description="Remove all polyhedras")

    def execute(self, context):
        obj = context.object
        batoms = Batoms(label=obj.batoms.label)
        index = batoms.coll.Bpolyhedra.ui_list_index
        batoms.polyhedra.settings.remove((self.species))
        batoms.coll.Bpolyhedra.ui_list_index = min(
            max(0, index - 1), len(batoms.polyhedra.settings) - 1
        )
        context.view_layer.objects.active = obj
        return {"FINISHED"}


class PolyhedraDraw(OperatorBatoms):
    bl_idname = "batoms.polyhedra_draw"
    bl_label = "Update polyhedra"
    bl_description = "Update polyhedra to a Batoms"

    def execute(self, context):
        obj = context.object
        batoms = Batoms(label=obj.batoms.label)
        batoms.model_style = "2"
        batoms.draw()
        context.view_layer.objects.active = batoms.obj
        return {"FINISHED"}


class PolyhedraModify(Operator):
    bl_idname = "batoms.polyhedra_modify"
    bl_label = "Modify polyhedra"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Modify polyhedra"

    key: StringProperty(
        name="key", default="style", description="Replaced by this polyhedra"
    )

    slice: BoolProperty(
        name="slice",
        default=False,
    )
    boundary: BoolProperty(
        name="boundary",
        default=False,
    )
    distance: FloatProperty(
        name="distance", description="Distance from origin", default=1
    )

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj:
            return obj.batoms.type == "MS" and obj.mode == "EDIT"
        else:
            return False

    def execute(self, context):
        obj = context.object
        data = obj.data
        bm = bmesh.from_edit_mesh(data)
        v = [s.index for s in bm.select_history if isinstance(s, bmesh.types.BMVert)]
        batoms = Batoms(label=obj.batoms.label)
        for i in v:
            setattr(batoms.bonds[i], self.key, getattr(self, self.key))
        # batoms.draw()
        return {"FINISHED"}



import bpy
from batoms.batoms import Batoms


def get_active_bpy_data(btype="cell"):
    """Helper function.

    Args:
        btype (str, optional): _description_. Defaults to 'cell'.
    """

    def getter():
        """Get the collection of the active Batoms

        When get the attribute of Batoms object,
        if the attribute if saved in the Batoms.coll.batoms,
        we only need to read data form the colleciton,
        it is faster than get data from the Batoms itself.

        Returns:
            bpy.type.collection: _description_
        """
        context = bpy.context
        if context.object and context.object.batoms.type != "OTHER":
            return getattr(bpy.data.collections[context.object.batoms.label], btype)
        return None

    return getter


def get_active_bpy_data_batoms(btype="cell"):
    """Helper function.

    Args:
        btype (str, optional): _description_. Defaults to 'cell'.
    """

    def getter():
        """

        Returns:
            bpy.type.collection: _description_
        """
        context = bpy.context
        if context.object and context.object.batoms.type != "OTHER":
            return getattr(
                bpy.data.collections[context.object.batoms.label].batoms, btype
            )
        return None

    return getter


def get_active_module(module_name="cell"):
    """Helper function.

    Args:
        module_name (str, optional): _description_. Defaults to 'cell'.

    Returns:
        function: function to get the module
    """

    def getter():
        """Get the module of Batoms by name

        Returns:
            _type_: _description_
        """
        context = bpy.context
        if context.object and context.object.batoms.type != "OTHER":
            mode = context.object.mode
            batoms = Batoms(label=context.object.batoms.label)
            bpy.context.view_layer.objects.active = batoms.obj
            bpy.ops.object.mode_set(mode=mode)
            return getattr(batoms, module_name)
        return None

    return getter


def set_module_attr(module_name):
    """Helper function.

    Args:
        module_name (_type_): _description_
    """

    def setter(key, value):
        module = get_active_module(module_name)()
        if module is not None:
            setattr(module, key, value)
            # bpy.context.view_layer.objects.active = module.obj

    return setter


def get_enum_attr(name, func):
    """Helper function to easily get enum property.

    Args:
        name (str): name of the attribute
    """

    def getter(self):
        batoms = func()
        if batoms is not None:
            return int(getattr(batoms, name))
        else:
            return 0

    return getter


def set_enum_attr(name, func):
    """Helper function to easily set enum property.

    Args:
        name (str): name of the attribute
    """

    def setter(self, value):
        items = self.bl_rna.properties[name].enum_items
        item = items[value]
        identifier = item.identifier
        self[name] = identifier
        func(name, value)

    return setter


def get_attr(name, func):
    """Helper function to easily get property.

    Args:
        name (str): name of the attribute
    """

    def getter(self):
        batoms = func()
        if batoms is not None:
            return getattr(batoms, name)
        else:
            prop = self.bl_rna.properties[name]
            return prop.default

    return getter


def set_attr(name, func):
    """Helper function to easily set property.

    Args:
        name (str): name of the attribute
    """

    def setter(self, value):
        # set value
        self[name] = value
        # set property of the module
        func(name, value)

    return setter



from .magres import Magres

from . import (
    bpy_data,
    ops,
    ui_list,
)

__all__ = ["Magres"]

classes = [
    bpy_data.MagresSetting,
    bpy_data.Magres,
    ops.MagresAdd,
    ops.MagresRemove,
    ops.MagresDraw,
    ui_list.BATOMS_MT_magres_context_menu,
    ui_list.BATOMS_UL_magres,
    ui_list.BATOMS_PT_magres,
]


def register_class():
    from bpy.types import Collection, Object
    from bpy.props import PointerProperty
    from bpy.utils import register_class

    for cls in classes:
        register_class(cls)
    # attach to blender internal data
    Collection.Bmagres = PointerProperty(name="Bmagres", type=bpy_data.Magres)
    Object.Bmagres = PointerProperty(name="Bmagres", type=bpy_data.MagresSetting)


def unregister_class():
    from bpy.types import Collection, Object
    from bpy.utils import unregister_class

    for cls in reversed(classes):
        unregister_class(cls)

    del Collection.Bmagres
    del Object.Bmagres



"""
"""
import bpy
from bpy.types import Panel, Operator
from bpy.props import StringProperty


class BatomsPropertiesIO(bpy.types.PropertyGroup):
    pubchem_cid: StringProperty(
        name="pubchem_cid", default="31423", description="pubchem_cid"
    )

    mp_key: StringProperty(name="mp_Key", default="", description="mp_key")

    mp_id: StringProperty(name="mp_ID", default="mp-2815", description="mp_id")

    rscb_name: StringProperty(name="rscb_name", default="1ema", description="rscb_name")


class Batoms_IO_search_material_project(Operator):
    bl_idname = "batoms.search_materials_project"
    bl_label = "Search"
    bl_description = "Search structure by id"

    def execute(self, context):
        from batoms.database.mp import mp_search

        io = context.scene.batoms.io
        mp_search(io.mp_key, io.mp_id)
        return {"FINISHED"}


class VIEW3D_PT_Batoms_io_materials_project(Panel):
    bl_label = "Materials Project"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"DEFAULT_CLOSED"}
    bl_category = "Batoms"
    bl_idname = "VIEW3D_PT_Batoms_io_materials_project"
    bl_parent_id = "VIEW3D_PT_Batoms_io"

    def draw(self, context):
        layout = self.layout
        io = context.scene.batoms.io

        layout.label(text="Search structure")
        layout.prop(io, "mp_key")
        layout.prop(io, "mp_id")
        layout.operator("batoms.search_materials_project")


class Batoms_IO_search_pubchem(Operator):
    bl_idname = "batoms.search_pubchem"
    bl_label = "Search"
    bl_description = "Search structure by id"

    def execute(self, context):
        from batoms.database.pubchem import pubchem_search

        io = context.scene.batoms.io
        pubchem_search(io.pubchem_cid)
        return {"FINISHED"}


class VIEW3D_PT_Batoms_io_pubchem(Panel):
    bl_label = "Pubchem"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"DEFAULT_CLOSED"}
    bl_category = "Batoms"
    bl_idname = "VIEW3D_PT_Batoms_io_pubchem"
    bl_parent_id = "VIEW3D_PT_Batoms_io"

    def draw(self, context):
        layout = self.layout
        io = context.scene.batoms.io

        layout.label(text="Search structure")
        layout.prop(io, "pubchem_cid")
        layout.operator("batoms.search_pubchem")


class Batoms_IO_search_RSCB(Operator):
    bl_idname = "batoms.search_rscb"
    bl_label = "Import"
    bl_description = "Import pdb by name"

    def execute(self, context):
        from batoms.database.rscb import rscb_import

        io = context.scene.batoms.io
        batoms = rscb_import(io.rscb_name)
        batoms.ribbon.draw()
        return {"FINISHED"}


class VIEW3D_PT_Batoms_io_RSCB(Panel):
    bl_label = "RSCB"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"DEFAULT_CLOSED"}
    bl_category = "Batoms"
    bl_idname = "VIEW3D_PT_Batoms_io_RSCB"
    bl_parent_id = "VIEW3D_PT_Batoms_io"

    def draw(self, context):
        layout = self.layout
        io = context.scene.batoms.io

        layout.label(text="Import PDB")
        layout.prop(io, "rscb_name")
        layout.operator("batoms.search_rscb")


class VIEW3D_PT_Batoms_io(Panel):
    bl_label = "Import-Export"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"DEFAULT_CLOSED"}
    bl_category = "Batoms"
    bl_idname = "VIEW3D_PT_Batoms_io"

    def draw(self, context):
        layout = self.layout
        layout.label(text="Import")
        layout.operator("batoms.import")
        layout.label(text="Export")
        layout.operator("batoms.export")



from bpy.props import (
    StringProperty,
    BoolProperty,
    IntProperty,
    FloatProperty,
    EnumProperty,
    FloatVectorProperty,
    CollectionProperty,
)

from batoms.internal_data import Base


class BondSetting(Base):
    """ """

    flag: BoolProperty(name="flag", default=False)
    label: StringProperty(name="label", default="batoms")
    # select: StringProperty(name="select", default='all')
    species1: StringProperty(name="species1")
    species2: StringProperty(name="species2")
    species: StringProperty(name="species")
    min: FloatProperty(name="min", description="min", default=0.0)
    max: FloatProperty(name="max", description="max", default=3.0)
    search: IntProperty(name="search", default=0)
    polyhedra: BoolProperty(name="polyhedra", default=False)
    color1: FloatVectorProperty(
        name="color1", size=4, subtype="COLOR", min=0, max=1, default=(0, 0.2, 0.8, 1)
    )
    color2: FloatVectorProperty(
        name="color2", size=4, subtype="COLOR", min=0, max=1, default=(0.6, 0.2, 0, 1)
    )
    width: FloatProperty(name="width", default=0.10)
    scale: FloatProperty(name="scale", default=1.0)
    order: IntProperty(name="order", default=1)
    segments: IntProperty(name="segments", default=16)
    order_offset: FloatProperty(name="order_offset", default=0.1)
    style: EnumProperty(
        name="style",
        description="bond style",
        items=(
            ("0", "Unicolor cylinder", ""),
            ("1", "Bicolor cylinder", ""),
            ("2", "Dashed line", ""),
            ("3", "Spring", "Spring"),
        ),
        default="1",
    )
    type: IntProperty(name="type", default=0)

    @property
    def name(self) -> str:
        return "%s-%s" % (self.species1, self.species2)

    def as_dict(self, reversed=False) -> dict:
        setdict = {
            "flag": self.flag,
            "label": self.label,
            "species1": self.species1,
            "species2": self.species2,
            "species": self.species,
            "name": self.name,
            "min": self.min,
            "max": self.max,
            "search": self.search,
            "polyhedra": self.polyhedra,
            "material_style": self.material_style,
            "color1": self.color1[:],
            "color2": self.color2[:],
            "width": self.width,
            "order": self.order,
            "order_offset": self.order_offset,
            "style": self.style,
            "type": self.type,
        }
        if reversed:
            setdict["name"] = "%s-%s" % (self.species2, self.species1)
            setdict["species1"] = self.species2
            setdict["species2"] = self.species1
            setdict["color1"] = self.color2[:]
            setdict["color2"] = self.color1[:]
        return setdict

    def __repr__(self) -> str:
        s = "-" * 60 + "\n"
        s = "Bondpair   min     max   Search_bond    Polyhedra \n"
        s += "{:10s} {:4.3f}   {:4.3f}      {:10s}   {:10s} \n".format(
            self.name, self.min, self.max, str(self.search), str(self.polyhedra)
        )
        s += "-" * 60 + "\n"
        return s


class Bond(Base):
    flag: BoolProperty(name="flag", default=False)
    label: StringProperty(name="label", default="batoms")
    show_search: BoolProperty(name="show_search", default=False)
    show_hydrogen_bond: BoolProperty(name="show_hydrogen_bond", default=False)
    ui_list_index: IntProperty(name="ui_list_index", default=0)
    # collection
    settings: CollectionProperty(name="bondsetting", type=BondSetting)



"""
Camera setting

"""
import bpy
from mathutils import Vector
from batoms.base.object import BaseObject
import logging

# logger = logging.getLogger('batoms')
logger = logging.getLogger(__name__)


def set_lock_camera_to_view(switch):
    for area in bpy.context.screen.areas:
        if area.type == "VIEW_3D":
            for space in area.spaces:
                if space.type == "VIEW_3D":
                    space.lock_camera = switch


class Camera(BaseObject):
    def __init__(
        self,
        label,
        name="Default",
        coll=None,
        type="ORTHO",
        lens=39.6,
        location=(0, 0, 30),
        look_at=None,
        lock_camera_to_view=True,
    ) -> None:
        """Camera Class

        Args:
            label (_type_): _description_
            name (str, optional): _description_. Defaults to 'Default'.
            coll (_type_, optional): _description_. Defaults to None.
            type (str, optional): _description_. Defaults to 'ORTHO'. 'ORTHO' or 'PERSP'
            lens (float, optional): _description_. Defaults to 39.6.
            location (tuple, optional): _description_. Defaults to (0, 0, 30).
            look_at (_type_, optional): _description_. Defaults to None.
            lock_camera_to_view (bool, optional): _description_. Defaults to True.
        """
        self.label = label
        self.name = name
        obj_name = "%s_camera_%s" % (label, name)
        btype = "Bcamera"
        BaseObject.__init__(self, obj_name=obj_name, btype=btype)
        self.create_camera(
            type,
            location,
            lens=lens,
            coll=coll,
            look_at=look_at,
            lock_camera_to_view=lock_camera_to_view,
        )
        self.obj.data.lens_unit = "FOV"

    def get_bpy_data(self):
        bpy_data = getattr(self.obj, self.btype)
        return bpy_data

    @property
    def lens(self):
        return self.get_lens()

    @lens.setter
    def lens(self, lens):
        self.set_lens(lens)

    def get_lens(self):
        return self.obj.data.lens

    def set_lens(self, lens):
        self.obj.data.lens = lens

    @property
    def ortho_scale(self):
        return self.get_ortho_scale()

    @ortho_scale.setter
    def ortho_scale(self, ortho_scale):
        self.set_ortho_scale(ortho_scale)

    def get_ortho_scale(self):
        return self.obj.data.ortho_scale

    def set_ortho_scale(self, ortho_scale):
        self.obj.data.ortho_scale = ortho_scale

    def create_camera(
        self,
        type=None,
        location=None,
        lens=50,
        coll=None,
        look_at=None,
        lock_camera_to_view=False,
    ):
        """
        type: str
            * PERSP
            * ORTHO
        lens: float
        """
        if self.obj_name in bpy.data.cameras:
            camera_data = bpy.data.cameras[self.obj_name]
        else:
            camera_data = bpy.data.cameras.new(self.obj_name)
        camera = bpy.data.objects.get(self.obj_name)
        # create camera if not exist
        if not camera:
            # bpy.data.objects.remove(obj, do_unlink=True)
            camera = bpy.data.objects.new(self.obj_name, camera_data)
            camera.data.lens = lens
            camera.data.type = type
            camera.location = Vector(location)
            if coll is not None:
                coll.objects.link(camera)
            set_lock_camera_to_view(lock_camera_to_view)
            if look_at is not None:
                self.look_at = look_at

    def __repr__(self) -> str:
        s = "Camera:   type = %s, location = %s, ortho_scale = %s \n" % (
            self.type,
            self.location,
            self.ortho_scale,
        )
        return s



"""Definition of the CrystalShape object in the Batoms package.

"""

import bpy
import numpy as np
from batoms.plugins.base import PluginObject
from .setting import CrystalShapeSettings
from batoms.draw import draw_cylinder, draw_surface_from_vertices
import logging

logger = logging.getLogger(__name__)


class CrystalShape(PluginObject):
    def __init__(
        self,
        label=None,
        location=np.array([0, 0, 0]),
        batoms=None,
    ):
        """CrystalShape Class

        Args:
            label (_type_, optional): _description_. Defaults to None.
            location (_type_, optional): _description_. Defaults to np.array([0, 0, 0]).
            batoms (_type_, optional): _description_. Defaults to None.
        """
        #
        self.batoms = batoms
        self.label = label
        PluginObject.__init__(self, label, "crystalshape", batoms)
        self.settings = CrystalShapeSettings(self.label, parent=self)
        self.settings.bpy_data.active = True

    def build_materials(self, name, color, node_inputs=None, material_style="default"):
        """ """
        from batoms.material import create_material

        if name in bpy.data.materials:
            mat = bpy.data.materials.get(name)
            bpy.data.materials.remove(mat, do_unlink=True)
        mat = create_material(
            name,
            color=color,
            node_inputs=node_inputs,
            material_style=material_style,
            backface_culling=False,
        )
        return mat

    def build_crystal(self, bcell, origin=[0, 0, 0]):
        """
        Build crystal

        no: Int
            spacegroup number
        """
        self.settings.get_symmetry_indices()
        planes = {}
        for p in self.settings:
            normal = np.dot(list(p.indices), bcell.reciprocal)
            normal = normal / np.linalg.norm(normal)
            point = p.distance * normal
            planes[p.name] = {
                "indices": list(p.indices),
                "normal": normal,
                "point": point,
                "vertices": [],
            }
        # loop all planes, find the intersection point of three plane
        keys = list(planes.keys())
        n = len(keys)
        vertices = []
        for i in range(n):
            for j in range(i + 1, n):
                for k in range(j + 1, n):
                    point = threePlaneIntersection(
                        [planes[keys[i]], planes[keys[j]], planes[keys[k]]]
                    )
                    # remove point outside plane
                    point = convexhull(planes, point)
                    if point is not None:
                        planes[keys[i]]["vertices"].append(point)
                        planes[keys[j]]["vertices"].append(point)
                        planes[keys[k]]["vertices"].append(point)
        new_planes = {}
        for name, plane in planes.items():
            p = self.settings[plane["indices"]]
            vertices, edges, faces = faces_from_vertices(
                plane["vertices"], plane["normal"]
            )
            if len(vertices) >= 3:
                vertices += np.array(origin)
                new_planes[p.name] = self.get_plane_data(vertices, edges, faces, p)
        self.crystal_planes = new_planes
        return new_planes

    def get_plane_data(self, vertices, edges, faces, p):
        """
        build edge
        """
        plane = {}
        if len(faces) > 0:
            plane = {
                "vertices": vertices,
                "edges": edges,
                "faces": faces,
                "material_style": p.material_style,
                "color": p.color,
                "indices": p.indices,
                "edges_cylinder": {
                    "lengths": [],
                    "centers": [],
                    "normals": [],
                    "vertices": 6,
                    "color": (0.0, 0.0, 0.0, 1.0),
                    "width": p.width,
                    "battr_inputs": {},
                },
                "battr_inputs": {"Bcrystalshape": p.as_dict()},
                "show_edge": p.show_edge,
                "slicing": p.slicing,
                "boundary": p.boundary,
            }
            for edge in edges:
                center = (vertices[edge[0]] + vertices[edge[1]]) / 2.0
                vec = vertices[edge[0]] - vertices[edge[1]]
                length = np.linalg.norm(vec)
                nvec = vec / length
                plane["edges_cylinder"]["lengths"].append(length)
                plane["edges_cylinder"]["centers"].append(center)
                plane["edges_cylinder"]["normals"].append(nvec)
        return plane

    def draw(self, plane_name="ALL", no=None, origin=None):
        """Draw crystal shape
        no: int
            spacegroup of structure, if None, no will be determined by
            get_spacegroup_number()
        origin: xyz vector
            The center of cyrstal shape
        """
        from batoms.utils.butils import clean_coll_object_by_type

        # delete old plane
        clean_coll_object_by_type(self.batoms.coll, "CRYSTALSHAPE")
        if no is not None:
            self.no = no
        if origin is None:
            origin = self.batoms.cell.origin
        planes = self.build_crystal(self.batoms.cell, origin=origin)
        for species, plane in planes.items():
            if plane_name.upper() != "ALL" and species != plane_name:
                continue
            name = "%s_%s_%s" % (self.label, self.name, species)
            self.delete_obj(name)
            obj = draw_surface_from_vertices(
                name,
                plane,
                coll=self.batoms.coll,
            )
            mat = self.build_materials(
                name, color=plane["color"], material_style=plane["material_style"]
            )
            obj.data.materials.append(mat)
            obj.parent = self.batoms.obj
            obj.batoms.type = "CRYSTALSHAPE"
            obj.batoms.label = self.batoms.label
            if plane["show_edge"]:
                name = "%s_%s_%s" % (self.label, "crystal_edge", species)
                self.delete_obj(name)
                obj = draw_cylinder(
                    name=name,
                    datas=plane["edges_cylinder"],
                    coll=self.batoms.coll,
                )
                obj.parent = self.batoms.obj
                obj.batoms.type = "CRYSTALSHAPE"
                obj.batoms.label = self.batoms.label

    @property
    def setting(self):
        from batoms.utils import deprecated

        """setting object."""
        deprecated(
            '"setting" will be deprecated in the furture, please use "settings".'
        )
        return self.settings


def save_image(data, filename, interpolation="bicubic"):
    """ """
    import pylab as plt
    import numpy as np

    data = data.T
    size = np.shape(data)
    logger.debug("size: {}".format(size))
    fig = plt.figure(figsize=(size[1] / size[0] * 10, 10))
    ax = plt.Axes(fig, [0.0, 0.0, 1.0, 1.0])
    ax.set_axis_off()
    fig.add_axes(ax)
    ax.imshow(data, interpolation=interpolation)
    plt.savefig(filename, dpi=300)


def faces_from_vertices(vertices, normal, include_center=False, scale=[1, 1, 1]):
    """
    get faces from vertices
    """
    # remove duplicative point
    vertices = np.unique(vertices, axis=0)
    n = len(vertices)
    if n < 3:
        return vertices, [], []
    center = np.mean(vertices, axis=0)
    v1 = vertices[0] - center
    angles = [[0, 0]]
    normal = normal / (np.linalg.norm(normal) + 1e-6)
    for i in range(1, n):
        v2 = vertices[i] - center
        x = np.cross(v1, v2)
        c = np.sign(np.dot(x, normal))
        angle = np.arctan2(c, np.dot(v1, v2))
        angles.append([i, angle])
    # scale
    vec = vertices - center
    # length = np.linalg.norm(vec, axis = 1)
    # nvec = vec/length[:, None]
    vertices = center + np.array([scale]) * vec
    # search convex polyhedra
    angles = sorted(angles, key=lambda x: x[1])
    if not include_center:
        faces = [[a[0] for a in angles]]
        edges = [[angles[0][0], angles[-1][0]]]
        for i in range(0, n - 1):
            edges.append([angles[i][0], angles[i + 1][0]])
    else:
        # add center to vertices
        vertices = np.append(vertices, center.reshape(1, 3), axis=0)
        icenter = len(vertices) - 1
        faces = [[angles[0][0], angles[-1][0], icenter]]
        edges = [
            [angles[0][0], icenter],
            [angles[0][0], angles[-1][0]],
            [angles[-1][0], icenter],
        ]
        for i in range(0, n - 1):
            faces.append([angles[i][0], angles[i + 1][0], icenter])
            edges.extend(
                [
                    [angles[i][0], angles[i + 1][0]],
                    [angles[i][0], icenter],
                    [angles[i + 1][0], icenter],
                ]
            )
    return vertices, edges, faces


def linePlaneIntersection(line, normal, point):
    """
    3D Line Segment and Plane Intersection
    - Point
    - Line contained in plane
    - No intersection
    """
    d = np.dot(point, normal)
    normalLine = line[0] - line[1]
    a = np.dot(normalLine, normal)
    # No intersection or Line contained in plane
    if np.isclose(a, 0):
        return None
    # in same side
    b = np.dot(line, normal) - d
    if b[0] * b[1] > 0:
        return None
    # Point
    v = point - line[0]
    d = np.dot(v, normal) / a
    point = np.round(line[0] + normalLine * d, 6)
    return point


def threePlaneIntersection(planes):
    """
    3D three Planes Intersection
    """
    # Condition for a single point of intersection
    mat = np.concatenate([[plane["normal"]] for plane in planes], axis=0)
    det = np.linalg.det(mat)
    if np.isclose(det, 0):
        return None
    darray = np.array([np.dot(plane["point"], plane["normal"]) for plane in planes])
    point = np.zeros(3)
    for i in range(3):
        temp = mat.copy()
        temp[:, i] = darray
        point[i] = round(np.linalg.det(temp) / det, 6)
    return point


def convexhull(planes, point):
    """
    find points at the same side of origin for all planes
    Todo: '+' or '-'
    """
    if point is None:
        return None
    for name, plane in planes.items():
        x1 = np.dot(point, plane["normal"]) - np.dot(plane["point"], plane["normal"])
        x2 = np.dot(np.array([0, 0, 0]), plane["normal"]) - np.dot(
            plane["point"], plane["normal"]
        )
        if abs(x1) > 1e-6 and x1 * x2 < -1e-6:
            return None
    return point

    def as_dict(self):
        """ """
        data = {}
        data["settings"] = self.settings.as_dict()
        data.update(self.settings.bpy_data.as_dict())
        return data



import sys
import subprocess
import importlib
from time import time
from bpy.types import Operator
from bpy.props import (
    StringProperty,
)
import logging

logger = logging.getLogger(__name__)

dependencies = {
    "ase": "ase",
    "scikit-image": "skimage",
}


def has_pip():
    # check pip exist or not
    return not subprocess.call([sys.executable, "-m", "pip", "--version"])


def has_module(modname):
    try:
        importlib.import_module(modname)
        return True
    except ImportError:
        return False


def install_pip():
    # if not exist
    cmd = [sys.executable, "-m", "ensurepip", "--upgrade"]
    subprocess.call(cmd)
    # upgrade
    cmd = [sys.executable, "-m", "pip", "install", "--upgrade", "pip"]
    subprocess.call(cmd)
    if not has_pip():
        logger.warning("pip cannot be installed, please install pip manually.")


def install_wheel():
    cmd = [sys.executable, "-m", "pip", "install", "wheel"]
    not subprocess.call(cmd)


def install_module(package, modname):
    if not has_module(modname):
        if not has_pip():
            install_pip()
        cmd = [sys.executable, "-m", "pip", "install", "--upgrade", package]
        subprocess.call(cmd)
        logger.info("package {0} installed.".format(package))
    # else:
    # print("package {0} installed.".format(package))
    return has_module(modname)


def install():
    tstart = time()
    for package, modname in dependencies.items():
        install_module(package, modname)
    logger.debug("Pip install time: {:.2f}".format(time() - tstart))


class BatomsInstallPackage(Operator):
    bl_idname = "batoms.pip_install_package"
    bl_label = "pip install"
    bl_description = "Install pip dependencies"

    package: StringProperty(name="package", default="ase", description="package name")

    modname: StringProperty(name="modname", default="ase", description="module name")

    def execute(self, context):
        if not has_pip():
            install_pip()
        install_module(self.package, self.modname)
        if not has_module(self.modname):
            self.report({"WARNING"}, "Cannot install package: {}".format(self.package))
            return {"CANCELLED"}
        self.report({"INFO"}, "Install package {} successfully!".format(self.package))
        return {"FINISHED"}


classes = [
    BatomsInstallPackage,
]


def register_class():
    from bpy.utils import register_class

    for cls in classes:
        register_class(cls)


def unregister_class():
    from bpy.utils import unregister_class

    for cls in reversed(classes):
        unregister_class(cls)



from bpy.props import (
    PointerProperty,
)
from . import bpy_data
from bpy.types import Collection, Object
from .bpy_data import Base

classes = [
    bpy_data.Belement,
    bpy_data.Bspecies,
    bpy_data.Batom,
    bpy_data.Battribute,
    bpy_data.Bcell,
    bpy_data.Bvolumetric_data,
    bpy_data.Bselect,
    bpy_data.Bboundary,
    bpy_data.BatomsCollection,
    bpy_data.BatomsObject,
]


__all__ = ["register_class", "unregister_class", "Base"]


def register_class():
    from bpy.utils import register_class

    for cls in classes:
        register_class(cls)

    Collection.batoms = PointerProperty(name="Batoms", type=bpy_data.BatomsCollection)
    Object.batoms = PointerProperty(name="Batoms", type=bpy_data.BatomsObject)


def unregister_class():
    from bpy.utils import unregister_class

    for cls in reversed(classes):
        unregister_class(cls)

    del Collection.batoms
    del Object.batoms



import bpy


def test_polyhedra(ch4):
    bpy.context.view_layer.objects.active = ch4.obj
    assert len(ch4.polyhedra.settings) == 2
    bpy.ops.batoms.polyhedra_draw()
    bpy.ops.batoms.polyhedra_remove(species="C")
    assert len(ch4.polyhedra.settings) == 1
    bpy.ops.batoms.polyhedra_add(species="C")
    assert len(ch4.polyhedra.settings) == 2



import bpy
from batoms import Batoms
import numpy as np
import pytest

if bpy.app.version_string >= "4.0.0":
    blender40 = False
else:
    blender40 = True


def test_empty():
    """Create an empty Batoms object"""
    from batoms import Batoms

    bpy.ops.batoms.delete()
    h2o = Batoms("h2o")
    assert len(h2o) == 0


def test_batoms_molecule():
    """Create a Batoms object from scratch"""
    bpy.ops.batoms.delete()
    from batoms import Batoms

    h2o = Batoms(
        "h2o",
        species=["O", "H", "H"],
        positions=[[0, 0, 0.40], [0, -0.76, -0.2], [0, 0.76, -0.2]],
    )
    assert len(h2o) == 3


def test_batoms_crystal():
    """Create a Batoms object with cell"""
    bpy.ops.batoms.delete()
    a = 4.08
    positions = [[0, 0, 0], [a / 2, a / 2, 0], [a / 2, 0, a / 2], [0, a / 2, a / 2]]
    au = Batoms(
        label="au",
        species=["Au"] * len(positions),
        positions=positions,
        pbc=True,
        cell=(a, a, a),
    )
    assert au.pbc == [True, True, True]
    assert np.isclose(au.cell[0, 0], a)


def test_batoms_parameters(h2o):
    """Create a Batoms object from scratch"""
    h2o.model_style = 1
    h2o.scale = 0.5
    assert h2o.model_style == 1
    assert np.isclose(h2o.scale, 0.5)
    # segments
    assert h2o.segments[0] == 24
    h2o.segments = [6, 6]
    assert h2o.segments[0] == 6


def test_model_style(tio2):
    tio2.boundary = 0.1
    tio2.bond.show_search = True
    tio2.model_style = 2
    assert tio2.bond.search_bond.hide is False
    assert tio2.boundary.hide is False
    #
    tio2.polyhedra_style = 2
    assert tio2.bond.hide is True
    assert tio2.bond.search_bond.hide is True
    assert tio2.boundary.hide is False
    #
    tio2.polyhedra_style = 1
    assert tio2.bond.hide is True
    assert tio2.bond.search_bond.hide is False
    assert tio2.boundary.hide is False


def test_batoms_write(h2o):
    """Export Batoms to structure file"""
    from ase.io import read

    h2o.write("h2o.in")
    new_h2o = read("h2o.in")
    assert len(h2o) == len(new_h2o)


def test_batoms_transform(h2o):
    """Transform: translate, rotate, mirror"""
    positions = h2o.positions
    h2o.translate([0, 0, 2])
    # local positions
    assert np.isclose(h2o.positions[0], positions[0]).all()
    # global positions
    assert np.isclose(h2o.global_positions[0], positions[0] + np.array([0, 0, 2])).all()
    # h2o.mirror("Z")
    # assert np.isclose(h2o.global_positions[0], np.array([0, 0, 1.6])).all()
    # h2o.rotate(90, "Z")
    # assert np.isclose(h2o.positions[0], np.array([0.76, 0, -2.2])).all()


def test_batoms_wrap(au):
    """Wrap atoms into pbc cell"""
    assert np.isclose(au[0].position, np.array([0, 0, 0])).all()
    au[0].position = np.array([-0.3, 0, 0])
    # evaluated
    # assert np.isclose(au[0].position, np.array([0.7, 0, 0])).all()


def test_batoms_supercell(au):
    """make supercell"""
    # repeat
    au = au * [2, 2, 2]
    assert len(au) == 32
    # transform
    P = np.array([[2, 3, 0, 5], [0, 1, 0, 5], [0, 0, 1, 0], [0, 0, 0, 1]])
    au = au.transform(P)
    # assert au.cell
    assert len(au) == 64


def test_batoms_occupy(h2o):
    """setting occupancy"""
    h2o.species["O"] = {"elements": {"O": 0.3, "Cl": 0.3}}


def test_batoms_copy(h2o):
    """copy batoms"""
    h2o.pbc = True
    h2o.cell = [3, 3, 3]
    h2o_2 = h2o.copy("h2o_2")
    assert isinstance(h2o_2, Batoms)
    assert len(h2o_2) == 3


def test_replace(h2o):
    """replace"""
    h2o.replace([1], "C")
    assert len(h2o.species) == 3
    assert h2o[1].species == "C"


def test_batoms_add(h2o):
    """Merge two Batoms objects"""
    from batoms import Batoms
    from ase.build import molecule

    co = Batoms("co", from_ase=molecule("CO"))
    batoms = h2o + co
    assert len(batoms) == 5
    assert len(batoms.species) == 3


def test_from_batoms(h2o):
    """Load Batoms"""
    new_h2o = Batoms("h2o")
    assert isinstance(new_h2o, Batoms)
    assert len(new_h2o.species) == 2
    assert len(new_h2o) == 3


def test_set_arrays(h2o):
    """Set arrays and attributes"""
    h2o.show = [1, 0, 1]
    assert not h2o[1].show
    h2o.scale = 1
    h2o.set_attributes({"scale": np.array([0.3, 0.3, 0.3])})
    # attributes
    h2o.show = 1
    h2o.show = True
    # positions
    positions = h2o.positions
    assert len(positions) == 3
    h2o.positions = h2o.positions + 2
    assert h2o.positions[0][0] == 2
    #
    del h2o[[2]]
    assert len(h2o.arrays["positions"]) == 2


def test_array_attribute():
    from ase.build import bulk
    import numpy as np
    from batoms import Batoms

    # (nx2) array
    au = bulk("Au")
    array2d = np.zeros((len(au), 2))
    au.set_array("array2d", array2d)
    #
    # (nx3) array
    array3d = np.zeros((len(au), 3))
    au.set_array("array3d", array3d)
    au = Batoms("au", from_ase=au)
    au.get_attribute("array2d")
    au.get_attribute("array3d")
    bpy.ops.batoms.delete()


@pytest.mark.skipif(
    blender40,
    reason="Requires Blender >= 4.0.0",
)
def test_array_attribute_4d():
    from ase.build import bulk
    import numpy as np
    from batoms import Batoms

    # (nx4) array
    au = bulk("Au")
    array4d = np.zeros((len(au), 4))
    au.set_array("array4d", array4d)
    au = Batoms("au", from_ase=au)
    au.get_attribute("array4d")
    bpy.ops.batoms.delete()


###############################
# Patch from TT for Atoms.array
###############################
def test_set_arrays_precision():
    import numpy as np
    from batoms import Batoms
    from batoms.bio.bio import read
    import ase.io

    atoms_ase = ase.io.read("../tests/datas/ch4_int_flag.extxyz")
    # ASE treats additional I-field as np.int32,
    # but on most recent platforms default int is np.int64
    # this may cause error in loading Batoms
    assert atoms_ase.arrays["some_int_flag"].dtype in (np.int32, np.int64)
    assert atoms_ase.arrays["positions"].dtype == np.float64
    # Without patch following will throw error
    atoms_bl = read("../tests/datas/ch4_int_flag.extxyz")
    # Change the array to float32 should also work
    atoms_ase2 = atoms_ase.copy()
    atoms_ase2.arrays["some_int_flag"] = np.array(
        [1.0, 2.0, 2.0, 2.0, 2.0], dtype="float32"
    )
    assert atoms_ase2.get_array("some_int_flag").dtype == np.float32
    atoms_bl2 = Batoms("ch4", from_ase=atoms_ase2)
    # get_attribute will return np.int64 and np.float64 explicitly
    assert atoms_bl.attributes["some_int_flag"].dtype == np.int64
    assert atoms_bl2.attributes["some_int_flag"].dtype == np.float64


def test_repeat(h2o):
    """Repeat"""
    import numpy as np

    h2o.cell = [3, 3, 3]
    h2o.pbc = True
    h2o.repeat([2, 2, 2])
    assert len(h2o) == 24
    print(h2o.positions)
    assert np.isclose(h2o.positions[3] - h2o.positions[0], np.array([0, 0, 3])).all()


def test_get_geometry(h2o):
    """Test geometry"""
    angle = h2o.get_angle(1, 0, 2)
    assert np.isclose(angle, 103.9998)
    d = h2o.get_distances(0, 1)
    assert np.isclose(d, 0.96856)
    com = h2o.get_center_of_mass()
    assert np.isclose(com, np.array([0, 0, 0.052531])).all()
    cog = h2o.get_center_of_geometry()
    assert np.isclose(cog, np.array([0, 0, -0.178892])).all()


def test_make_real(h2o):
    h2o.realize_instances = True
    h2o.realize_instances = False


def test_as_dict(au):
    data = au.as_dict()
    assert "lattice_plane" not in data.keys()
    # active plugin
    au.lattice_plane
    data = au.as_dict()
    assert "lattice_plane" in data.keys()


def test_batoms_ops():
    from batoms import Batoms

    bpy.ops.batoms.delete()
    bpy.ops.batoms.bulk_add(label="au", formula="Au")
    au = Batoms("au")
    assert len(au) == 1


def test_export_mesh_x3d(c2h6so):
    from batoms.bio.bio import read

    c2h6so.cell = [3, 3, 3]
    c2h6so.model_style = 1
    c2h6so.export_mesh("c2h6so.x3d", with_cell=True, with_bond=True)
    bpy.ops.batoms.delete()
    tio2 = read("../tests/datas/tio2.cif")
    tio2.boundary = 0.01
    tio2.model_style = 2
    tio2.bond.show_search = True
    tio2.export_mesh(
        "tio2.x3d",
        with_cell=True,
        with_polyhedra=True,
        with_boundary=True,
        with_search_bond=True,
        with_bond=True,
    )


@pytest.mark.skipif(
    not blender40, reason="In Blender 4.0, export_scene.obj is not working."
)
def test_export_mesh_obj(c2h6so, tio2):
    c2h6so.cell = [3, 3, 3]
    c2h6so.model_style = 1
    c2h6so.export_mesh("c2h6so.obj", with_cell=True, with_bond=True)
    tio2.boundary = 0.01
    tio2.model_style = 2
    tio2.bond.show_search = True
    tio2.export_mesh(
        "tio2.obj",
        with_cell=True,
        with_polyhedra=True,
        with_boundary=True,
        with_search_bond=True,
        with_bond=True,
    )


def test_calc_electrostatic_potential(c2h6so):
    from time import time

    # pytest skip if openbabel is not installed
    pytest.importorskip("openbabel")
    tstart = time()
    c2h6so.auto_assign_charge()
    t = time() - tstart
    print("auto_assign_charge: {:1.2f}".format(t))
    tstart = time()
    c2h6so.calc_electrostatic_potential(c2h6so.positions - 1)
    t = time() - tstart
    print("calc_electrostatic_potential: {:1.2f}".format(t))
    assert t < 5



import bpy
import console_python
import logging

logger = logging.getLogger(__name__)


def console_hook():
    """add batoms to namespace of python console"""
    from batoms.utils.butils import read_batoms_list

    items = read_batoms_list()
    for area in bpy.context.screen.areas:
        if area.type == "CONSOLE":
            for region in area.regions:
                if region.type == "WINDOW":
                    console, stdout, stderr = console_python.get_console(hash(region))
                    if "Batoms" not in console.locals:
                        console.push("from batoms import Batoms")
                    for item in items:
                        item = item.replace("-", "_")
                        item = item.replace(".", "")
                        if item[:1].isdigit():
                            item = "b_" + item
                        if item not in console.locals:
                            logger.info("Python console: Add Batoms {}".format(item))
                            console.push("{}=Batoms('{}')".format(item, item))
                            console.push("")
                            # console.locals[item] = Batoms(item)


def register_hook():
    console_python.execute.hooks.append((console_hook, ()))


def unregister_hook():
    console_python.execute.hooks.remove((console_hook, ()))



import bpy


def test_magres():
    from batoms import Batoms
    from ase.io import read

    bpy.ops.batoms.delete()
    atoms = read("../tests/datas/ethanol.magres")
    ms_array = atoms.arrays.pop("ms")
    print(ms_array.shape)
    for i in range(3):
        for j in range(3):
            atoms.set_array("ms_%d%d" % (i, j), ms_array[:, i, j])

    efg_array = atoms.arrays.pop("efg")
    for i in range(3):
        for j in range(3):
            atoms.set_array("efg_%d%d" % (i, j), efg_array[:, i, j])

    ethanol = Batoms("ethanol", from_ase=atoms)
    ethanol.model_style = 1
    ethanol.magres.settings["1"].scale = 0.005
    ethanol.magres.draw()


def test_magres_uilist():
    """magres panel"""
    from batoms import Batoms
    from ase.io import read

    bpy.ops.batoms.delete()
    atoms = read("../tests/datas/ethanol.magres")
    ethanol = Batoms("ethanol", from_ase=atoms)
    bpy.context.view_layer.objects.active = ethanol.obj
    assert ethanol.coll.Bmagres.ui_list_index == 0
    bpy.ops.surface.magres_add(name="2")
    assert ethanol.coll.Bmagres.ui_list_index == 1



import urllib.request
from batoms import Batoms
from batoms.pdbparser import read_pdb
import logging

# logger = logging.getLogger('batoms')
logger = logging.getLogger(__name__)


def rscb_import(name):
    if not name.endswith(".pdb"):
        name += ".pdb"
    urllib.request.urlretrieve("http://files.rcsb.org/download/%s" % name, name)
    atoms = read_pdb(name)
    batoms = Batoms("pdb_%s" % name[0:-4], from_ase=atoms)
    batoms.ribbon.draw()
    return batoms


if __name__ == "__main__":
    rscb_import("1ema")



import bpy
from ase.build import bulk
from batoms import Batoms
import os

try:
    from _common_helpers import has_display

    use_cycles = not has_display()
except ImportError:
    use_cycles = False

extras = dict(engine="cycles") if use_cycles else {}

skip_test = bool(os.environ.get("NOTEST_CUBE", 0))


def test_template():
    bpy.ops.batoms.delete()
    au = bulk("Au", cubic=True)
    au = Batoms("au", from_ase=au)
    au.template.settings["test"] = {"prop1": 1.1}
    print(au.template.settings)
    au.template.draw()


def test_template_ops():
    bpy.ops.batoms.delete()
    bpy.ops.batoms.bulk_add(formula="Au")
    au = Batoms("Au")
    bpy.context.view_layer.objects.active = au.obj
    bpy.ops.template.template_add(name="test")
    au.template.settings["test"].scale = 3.0
    assert len(au.template.settings) == 1
    bpy.ops.template.template_draw()
    bpy.ops.template.template_remove(name="test")
    assert len(au.template.settings) == 0
    print(au.template.settings)
    bpy.ops.template.template_draw()


def test_gui():
    """latticeplane panel"""
    from batoms.batoms import Batoms

    bpy.ops.batoms.delete()
    bpy.ops.batoms.bulk_add(formula="Au", cubic=True)
    au = Batoms("Au")
    au.obj.select_set(True)
    assert bpy.context.scene.Btemplate.show is True
    bpy.context.scene.Btemplate.show = False
    assert au.template.show is False


def test_gui_uilist():
    """latticeplane panel"""
    from batoms.batoms import Batoms

    bpy.ops.batoms.delete()
    bpy.ops.batoms.bulk_add(formula="Au", cubic=True)
    au = Batoms("Au")
    au.obj.select_set(True)
    assert au.coll.Btemplate.ui_list_index == 0
    bpy.ops.template.template_add(name="test1")
    bpy.ops.template.template_add(name="test2")
    assert au.coll.Btemplate.ui_list_index == 1


if __name__ == "__main__":
    test_template()
    print("\n Template: All pass! \n")
